"""
 @file
 @brief This file is for legacy support of OpenShot 1.x project files
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import uuid


class track:
    """The track class contains a simple grouping of clips on the same layer (aka track)."""

    # ----------------------------------------------------------------------
    def __init__(self, track_name, parent_sequence):
        """Constructor"""

        # init variables for sequence
        self.name = track_name
        self.x = 10  # left x coordinate to start the track
        self.y_top = 0  # top y coordinate of this track
        self.y_bottom = 0  # bottom y coordinate of this track
        self.parent = parent_sequence  # reference to parent sequence object
        self.play_video = True
        self.play_audio = True
        self.unique_id = str(uuid.uuid1())

        # init the tracks on the sequence
        self.clips = []

        # init transitions
        self.transitions = []



"""
 @file
 @brief This file loads the Choose Profile dialog
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2023 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import functools

import openshot  # Python module for libopenshot (required video editing module installed separately)

from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QDialog, QSizePolicy, QDialogButtonBox

from classes import info, ui_util, tabstops
from classes.app import get_app
from classes.logger import log
from classes.metrics import track_metric_screen
from windows.views.profiles_treeview import ProfilesTreeView


class Profile(QDialog):
    """ Choose Profile Dialog """

    # Path to ui file
    ui_path = os.path.join(info.PATH, 'windows', 'ui', 'profile.ui')

    def __init__(self, initial_profile_desc=None):

        # Create dialog class
        QDialog.__init__(self)

        # Load UI from designer & init
        ui_util.load_ui(self, self.ui_path)
        ui_util.init_ui(self)

        # get translations
        _ = get_app()._tr

        # Pause playback
        get_app().window.PauseSignal.emit()

        # Track metrics
        track_metric_screen("profile-screen")

        # Keep track of starting selection
        self.initial_index = 0

        # Set up the buttons
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.okButton = self.buttonBox.button(QDialogButtonBox.Ok)
        self.cancelButton = self.buttonBox.button(QDialogButtonBox.Cancel)

        # Set object names (for theme styles)
        self.okButton.setObjectName("acceptButton")
        self.cancelButton.setObjectName("cancelButton")
        self.layout().addWidget(self.buttonBox)

        # Connect the buttons
        self.okButton.clicked.connect(self.accept)
        self.cancelButton.clicked.connect(self.reject)

        # Loop through profiles
        self.profile_list = []
        self.project_profile = None
        self.selected_profile = None
        self.project_index = 0
        for profile_folder in [info.USER_PROFILES_PATH, info.PROFILES_PATH]:
            for file in reversed(sorted(os.listdir(profile_folder))):
                profile_path = os.path.join(profile_folder, file)
                if os.path.isdir(profile_path):
                    continue
                try:
                    # Load Profile
                    profile = openshot.Profile(profile_path)
                    if profile_folder == info.USER_PROFILES_PATH:
                        profile.path = profile_path
                        profile.user_created = True
                    else:
                        profile.user_created = False
                    if profile.info.description == initial_profile_desc or profile.Key() == initial_profile_desc:
                        self.project_profile = profile
                        self.project_index = len(self.profile_list)
                        self.setWindowTitle(f'{_("Choose Profile")} [{profile.info.description}]')

                    # Add description of Profile to list
                    self.profile_list.append(profile)

                except RuntimeError as e:
                    log.warning("Failed to parse file '%s' as a profile: %s" % (profile_path, e))

        # Create treeview
        self.profileListView = ProfilesTreeView(self, self.profile_list)
        self.profileListView.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        self.verticalLayout.insertWidget(1, self.profileListView)

        # Select project profile (if any)
        if self.project_profile:
            model_index = self.profileListView.profiles_model.proxy_model.index(self.project_index, 1)
            self.profileListView.select_profile(model_index)

        # Connect signals
        self.txtProfileFilter.textChanged.connect(self.profileListView.refresh_view)
        self.txtProfileFilter.textChanged.connect(self.profileListView.refresh_view)
        self.profileListView.FilterCountChanged.connect(self.profileCountChanged)
        self.profileListView.doubleClicked.connect(self.profileDoubleClick)

        tabstops.apply_auto_tab_order_later(self)

    def profileCountChanged(self, new_count):
        """Profile filter count changed"""
        self.lblCount.setText(f"{new_count}")

    def profileDoubleClick(self):
        """Profile tree was double clicked"""
        self.accept()

    def accept(self):
        """ Ok button clicked """
        # Get selected profile
        profile = self.profileListView.get_profile()
        if profile:
            # New profile selected
            self.selected_profile = profile
            super(Profile, self).accept()
        else:
            # No profile or same as current project
            self.reject()

    def closeEvent(self, event):
        """Signal for closing Profile window"""
        # Invoke the close button
        self.reject()

    def reject(self):
        """Window closed without choosing a new profile"""
        # Close dialog
        super(Profile, self).reject()



"""
 @file
 @brief This file loads the current language based on the computer's locale settings
 @author Noah Figg <eggmunkee@hotmail.com>
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import locale

from PyQt5.QtCore import QLocale, QLibraryInfo, QTranslator, QCoreApplication

from classes.logger import log
from classes import info

try:
    from language import openshot_lang  # noqa
    language_path = ":/locale/"
    log.debug("Using compiled translation resources")
except ImportError:
    language_path = os.path.join(info.PATH, 'language')
    log.debug("Loading translations from: {}".format(language_path))


def init_language():
    """ Find the current locale, and install the correct translators """

    # Get app instance
    app = QCoreApplication.instance()

    # Setup of our list of translators and paths
    translator_types = (
        {"type": 'QT',
         "prefix": 'qt_',        # Older versions of Qt use this file (built-in translations)
         "path": QLibraryInfo.location(QLibraryInfo.TranslationsPath)},
        {"type": 'QT',
         "prefix": 'qtbase_',    # Newer versions of Qt use this file (built-in translations)
         "path": QLibraryInfo.location(QLibraryInfo.TranslationsPath)},
        {"type": 'QT',
         "prefix": 'qt_',
         "path": os.path.join(info.PATH, 'language')}, # Optional path where we package QT translations
        {"type": 'QT',
         "prefix": 'qtbase_',
         "path": os.path.join(info.PATH, 'language')}, # Optional path where we package QT translations
        {"type": 'OpenShot',
         "prefix": 'OpenShot_',  # Our custom translations
         "path": language_path},
    )

    # Determine the environment locale, or default to system locale name
    locale_names = [os.environ.get('LANG', QLocale().system().name()),
                    os.environ.get('LOCALE', QLocale().system().name())
                    ]

    # Get the user's configured language preference
    settings = app.get_settings()
    if settings:
        preference_lang = settings.get('default-language')
    else:
        preference_lang = "Default"

    # Output all languages detected from various sources
    log.info("Qt Detected Languages: {}".format(QLocale().system().uiLanguages()))
    log.info("LANG Environment Variable: {}".format(os.environ.get('LANG', "")))
    log.info("LOCALE Environment Variable: {}".format(os.environ.get('LOCALE', "")))
    log.info("OpenShot Preference Language: {}".format(preference_lang))

    # Check if the language preference is something other than "Default"
    if preference_lang == "en_US":
        # Override language list with en_US, don't add to it
        locale_names = [ "en_US" ]
    elif preference_lang != "Default":
        # Prepend preference setting to list
        locale_names.insert(0, preference_lang)

    # If the user has used the --lang command line arg, override with that
    # (We've already checked that it's in SUPPORTED_LANGUAGES)
    if info.CMDLINE_LANGUAGE:
        locale_names = [ info.CMDLINE_LANGUAGE ]
        log.info("Language overridden on command line, using: {}".format(info.CMDLINE_LANGUAGE))

    # Default the locale to C, for number formatting
    locale.setlocale(locale.LC_ALL, 'C')

    # Loop through environment variables
    found_language = False
    for locale_name in locale_names:

        # Go through each translator and try to add for current locale
        for type in translator_types:
            trans = QTranslator(app)
            if find_language_match(type["prefix"], type["path"], trans, locale_name):
                # Install translation
                app.installTranslator(trans)
                found_language = True

        # Exit if found language for type: "OpenShot"
        if found_language:
            log.debug("Exiting translation system (since we successfully loaded: {})".format(locale_name))
            info.CURRENT_LANGUAGE = locale_name
            break


# Try the full locale and base locale trying to find a valid path
#  returns True when a match was found.
#  pattern - a string expected to have one pipe to be filled by locale strings
#  path - base path for file (pattern may contain more path)
#
def find_language_match(prefix, path, translator, locale_name):
    """ Match all combinations of locale, language, and country """

    filename = prefix + locale_name
    log.debug('Attempting to load {} in \'{}\''.format(filename,path))
    success = translator.load(filename, path)
    if success:
        log.debug('Successfully loaded {} in \'{}\''.format(filename, path))
    return success


def get_all_languages():
    """Get all language names and countries packaged with OpenShot"""

    # Loop through all supported language locale codes
    all_languages = []
    for locale_name in info.SUPPORTED_LANGUAGES:
        try:
            native_lang_name = QLocale(locale_name).nativeLanguageName().title()
            country_name = QLocale(locale_name).nativeCountryName().title()
            all_languages.append((locale_name, native_lang_name, country_name))
        except Exception:
            log.debug('Failed to parse language for %s', locale_name)

    # Return list
    return all_languages


def get_current_locale():
    return info.CURRENT_LANGUAGE



"""
 @file
 @brief This file contains unit tests for JSON data loading and saving
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import sys
import tempfile
import unittest
from contextlib import ExitStack
from pathlib import Path
from unittest.mock import patch


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from PyQt5.QtWidgets import QApplication

from classes import info
from classes.json_data import JsonDataStore
from qt_test_app import ensure_app_state as ensure_qt_app_state, get_or_create_app


class DummySettings:
    def __init__(self):
        self.values = {
            "default-profile": "HD 720p 30 fps",
            "default-samplerate": 48000,
            "default-channels": 2,
        }

    def get(self, key):
        return self.values.get(key)

    def set(self, key, value):
        self.values[key] = value


class DummyApp(QApplication):
    def __init__(self):
        super().__init__([])
        self.settings = DummySettings()
        from classes.project_data import ProjectDataStore
        from classes.updates import UpdateManager

        self.project = ProjectDataStore()
        self.updates = UpdateManager()
        self.updates.add_listener(self.project)
        self.updates.reset()
        self.window = None

    def get_settings(self):
        return self.settings

    def _tr(self, text):
        return text


def ensure_app_state(app):
    from classes.project_data import ProjectDataStore
    from classes.updates import UpdateManager

    return ensure_qt_app_state(
        app,
        DummySettings,
        project_factory=ProjectDataStore,
        updates_factory=UpdateManager,
        extra_attrs={"window": None},
    )


class JsonDataTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        app, cls._owns_app = get_or_create_app(DummyApp)
        cls.app = ensure_app_state(app)

    @classmethod
    def tearDownClass(cls):
        if getattr(cls, "_owns_app", False) and cls.app:
            cls.app.quit()

    def test_read_from_file_repairs_windows_drive_corruption(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            project_file = Path(tmpdir) / "broken.osp"
            project_file.write_text(
                '{\n'
                ' "version": {"openshot-qt": "2.6.0"},\n'
                ' "files": [\n'
                '  {\n'
                '   id: "F1",\n'
                '   "path": "C:/media/clip.mp4"\n'
                '  }\n'
                ' ]\n'
                '}\n',
                encoding="utf-8",
            )

            store = JsonDataStore()
            data = store.read_from_file(str(project_file))

            self.assertEqual(data["files"][0]["id"], "F1")
            self.assertTrue(project_file.with_suffix(".osp.bak").exists())
            self.assertIn('"id": "F1"', project_file.read_text(encoding="utf-8"))

    def test_write_relative_and_read_absolute_round_trip(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            project_file = tmpdir / "project.osp"
            assets_dir = tmpdir / "project_assets"
            transitions_dir = tmpdir / "openshot" / "transitions"
            colors_dir = tmpdir / "colors"

            import classes.json_data as json_data_module

            with ExitStack() as stack:
                stack.enter_context(patch.object(info, "PATH", str(tmpdir / "openshot")))
                stack.enter_context(patch.object(info, "COLORS_PATH", str(colors_dir)))
                stack.enter_context(patch.object(info, "THUMBNAIL_PATH", str(tmpdir / "thumbs")))
                stack.enter_context(
                    patch.object(
                        json_data_module,
                        "get_assets_path",
                        lambda file_path, create_paths=False: str(assets_dir),
                    )
                )
                store = JsonDataStore()
                source = {
                    "path": str(assets_dir / "title" / "intro.svg"),
                    "resource": str(transitions_dir / "common" / "fade.svg"),
                    "lut_path": str(colors_dir / "teal.cube"),
                }

                store.write_to_file(str(project_file), source, path_mode="relative")
                saved = project_file.read_text(encoding="utf-8")

                self.assertIn('"path": "@assets/title/intro.svg"', saved)
                self.assertTrue(
                    '"resource": "@transitions/common/fade.svg"' in saved
                    or f'"resource": "{str(transitions_dir / "common" / "fade.svg")}"' in saved
                )
                self.assertIn('"lut_path": "@colors/teal.cube"', saved)

                loaded = store.read_from_file(str(project_file), path_mode="absolute")
                self.assertEqual(loaded["path"], str(assets_dir / "title" / "intro.svg"))
                self.assertEqual(loaded["resource"], str(transitions_dir / "common" / "fade.svg"))
                self.assertEqual(loaded["lut_path"], str(colors_dir / "teal.cube"))

    def test_make_repair_backup_increments_suffix_when_backup_exists(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            project_file = Path(tmpdir) / "project.osp"
            project_file.write_text("{}", encoding="utf-8")
            Path(f"{project_file}.bak").write_text("old", encoding="utf-8")

            store = JsonDataStore()
            store.make_repair_backup(str(project_file), '{"id": "P1"}')

            second_backup = Path(f"{project_file}.bak.1")
            self.assertTrue(second_backup.exists())
            self.assertEqual(second_backup.read_text(encoding="utf-8"), '{"id": "P1"}')



"""
 @file
 @brief This file contains the zoom slider QWidget (for interactive zooming/panning on the timeline)
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """
import copy
import math

from PyQt5.QtCore import (
    Qt, QCoreApplication, QRectF, QTimer, QSize
)
from PyQt5.QtGui import (
    QPainter, QColor, QPen, QBrush, QCursor, QPainterPath, QIcon
)
from PyQt5.QtWidgets import QSizePolicy, QWidget

import openshot  # Python module for libopenshot (required video editing module installed separately)

from classes import updates
from classes.app import get_app
from classes.query import Clip, Track, Transition, Marker
from classes.logger import log


class ZoomSlider(QWidget, updates.UpdateInterface):
    """ A QWidget used to zoom and pan around a Timeline"""

    def sizeHint(self):
        """Preferred size for layouts that host the slider."""
        return QSize(200, 20)

    def minimumSizeHint(self):
        """Allow the slider to shrink horizontally when space is limited."""
        return QSize(0, 20)

    # This method is invoked by the UpdateManager each time a change happens (i.e UpdateInterface)
    def changed(self, action):
        # Ignore changes that don't affect this
        if (action and len(action.key) >= 1 and action.key[0].lower() in ["files", "history", "profile"]) or self.ignore_updates:
            return

        # Clear previous rects
        self.clip_rects.clear()
        self.clip_rects_selected.clear()
        self.marker_rects.clear()

        # Get layer lookup
        layers = {}
        for count, layer in enumerate(reversed(sorted(Track.filter()))):
            layers[layer.data.get('number')] = count

        # Wait for timeline object and valid scrollbar positions
        # TODO: Fix commented out logic
        if hasattr(get_app().window, "timeline"):  # and self.scrollbar_position[2] != 0.0:
            # Get max width of timeline
            project_duration = get_app().project.get("duration")
            pixels_per_second = self.width() / project_duration

            # Determine scale factor
            vertical_factor = self.height() / len(layers.keys())

            for clip in Clip.filter():
                # Calculate clip geometry (and cache it)
                clip_x = (clip.data.get('position', 0.0) * pixels_per_second)
                clip_y = layers.get(clip.data.get('layer', 0), 0) * vertical_factor
                clip_width = ((clip.data.get('end', 0.0) - clip.data.get('start', 0.0))
                              * pixels_per_second)
                clip_rect = QRectF(clip_x, clip_y, clip_width, 1.0 * vertical_factor)
                if clip.id in get_app().window.selected_clips:
                    # selected clip
                    self.clip_rects_selected.append(clip_rect)
                else:
                    # un-selected clip
                    self.clip_rects.append(clip_rect)

            for clip in Transition.filter():
                # Calculate clip geometry (and cache it)
                clip_x = (clip.data.get('position', 0.0) * pixels_per_second)
                clip_y = layers.get(clip.data.get('layer', 0), 0) * vertical_factor
                clip_width = ((clip.data.get('end', 0.0) - clip.data.get('start', 0.0))
                              * pixels_per_second)
                clip_rect = QRectF(clip_x, clip_y, clip_width, 1.0 * vertical_factor)
                if clip.id in get_app().window.selected_transitions:
                    # selected clip
                    self.clip_rects_selected.append(clip_rect)
                else:
                    # un-selected clip
                    self.clip_rects.append(clip_rect)

            for marker in Marker.filter():
                # Calculate clip geometry (and cache it)
                marker_x = (marker.data.get('position', 0.0) * pixels_per_second)
                marker_rect = QRectF(marker_x, 0, 0.5, len(layers) * vertical_factor)
                self.marker_rects.append(marker_rect)

        # Force re-paint
        self.update()

    def paintEvent(self, event, *args):
        """ Custom paint event """
        event.accept()

        # Get theme colors
        if get_app().theme_manager:
            theme = get_app().theme_manager.get_current_theme()
            if not theme:
                log.warning("No theme loaded yet. Skip rendering zoom slider widget.")
                return
            playhead_color = theme.get_color(".zoom_slider_playhead", "background-color")
        else:
            log.warning("No ThemeManager loaded yet. Skip rendering zoom slider widget.")

        # Paint timeline preview on QWidget
        painter = QPainter(self)
        painter.setRenderHints(QPainter.Antialiasing | QPainter.SmoothPixmapTransform | QPainter.TextAntialiasing, True)

        # Fill the whole widget with the solid color (background solid color)
        background_color = self.palette().color(self.palette().Base)
        painter.fillRect(event.rect(), background_color)

        # Create pens / colors
        clip_pen = QPen(QBrush(QColor("#53a0ed")), 1.5)
        clip_pen.setCosmetic(True)
        painter.setPen(clip_pen)

        selected_clip_pen = QPen(QBrush(QColor("Red")), 1.5)
        selected_clip_pen.setCosmetic(True)

        scroll_color = QColor("#4053a0ed")
        scroll_pen = QPen(QBrush(scroll_color), 2.0)
        scroll_pen.setCosmetic(True)

        marker_color = QColor("#4053a0ed")
        marker_pen = QPen(QBrush(marker_color), 1.0)
        marker_pen.setCosmetic(True)

        playhead_color = playhead_color
        playhead_pen = QPen(QBrush(playhead_color), 1.0)
        playhead_pen.setCosmetic(True)

        handle_color = QColor("#a653a0ed")
        handle_pen = QPen(QBrush(handle_color), 1.5)
        handle_pen.setCosmetic(True)

        # Get layer lookup
        layers = Track.filter()

        # Wait for timeline object and valid scrollbar positions
        if get_app().window.timeline:
            # Get max width of timeline
            project_duration = get_app().project.get("duration")
            pixels_per_second = event.rect().width() / project_duration
            scroll_width = (self.scrollbar_position[1] - self.scrollbar_position[0]) * event.rect().width()

            # Get FPS info
            fps_num = get_app().project.get("fps").get("num", 24)
            fps_den = get_app().project.get("fps").get("den", 1) or 1
            fps_float = float(fps_num / fps_den)

            # Determine scale factor
            vertical_factor = event.rect().height() / len(layers)

            # Loop through each clip
            painter.setPen(clip_pen)
            for clip_rect in self.clip_rects:
                painter.drawRect(clip_rect)

            painter.setPen(selected_clip_pen)
            for clip_rect in self.clip_rects_selected:
                painter.drawRect(clip_rect)

            painter.setPen(marker_pen)
            for marker_rect in self.marker_rects:
                painter.drawRect(marker_rect)

            painter.setPen(playhead_pen)
            playhead_x = ((self.current_frame / fps_float) * pixels_per_second)
            playhead_rect = QRectF(playhead_x, 0, 0.5, len(layers) * vertical_factor)
            painter.drawRect(playhead_rect)

            # Draw scroll bars (if available)
            if self.scrollbar_position:
                painter.setPen(scroll_pen)

                # scroll bar path
                scroll_x = self.scrollbar_position[0] * event.rect().width()
                self.scroll_bar_rect = QRectF(scroll_x, 0.0, scroll_width, event.rect().height())
                scroll_path = QPainterPath()
                scroll_path.addRoundedRect(self.scroll_bar_rect, 6, 6)

                # draw scroll bar rect
                painter.fillPath(scroll_path, scroll_color)
                painter.drawPath(scroll_path)

                # draw handles
                painter.setPen(handle_pen)
                handle_width = 12.0

                # left handle
                left_handle_x = (self.scrollbar_position[0] * event.rect().width()) - (handle_width/2.0)
                self.left_handle_rect = QRectF(left_handle_x, event.rect().height() / 4.0, handle_width, event.rect().height() / 2.0)
                left_handle_path = QPainterPath()
                left_handle_path.addRoundedRect(self.left_handle_rect, handle_width, handle_width)
                painter.fillPath(left_handle_path, handle_color)

                # right handle
                right_handle_x = self.scroll_bar_rect.right() - (handle_width/2.0)
                right_handle_x = min(right_handle_x, event.rect().width() - (handle_width/2.0))
                self.right_handle_rect = QRectF(right_handle_x, event.rect().height() / 4.0, handle_width, event.rect().height() / 2.0)
                right_handle_path = QPainterPath()
                right_handle_path.addRoundedRect(self.right_handle_rect, handle_width, handle_width)
                painter.fillPath(right_handle_path, handle_color)

            # Determine if play-head is inside scroll area
            if get_app().window.preview_thread.player.Mode() == openshot.PLAYBACK_PLAY and self.is_auto_center:
                if not self.scroll_bar_rect.contains(playhead_rect):
                    get_app().window.TimelineCenter.emit()

        # End painter
        painter.end()

    def zoomToTimeline(self):
        """Toggle between zooming to the entire timeline and the previous zoom"""
        # Are we already zoomed complete out?
        if math.isclose(self.scrollbar_position[0], 0.0, abs_tol=1e-9) and math.isclose(self.scrollbar_position[1], 1.0, abs_tol=1e-9):
            # Restore previous zoom
            self.scrollbar_position[0] = self.scrollbar_zoom_previous[0]
            self.scrollbar_position[1] = self.scrollbar_zoom_previous[1]
        else:
            # Zoom out to reveal the entire timeline
            self.scrollbar_zoom_previous = copy.deepcopy(self.scrollbar_position)
            self.scrollbar_position[0] = 0.0
            self.scrollbar_position[1] = 1.0
        self.delayed_resize_callback()

    def mouseDoubleClickEvent(self, event):
        self.zoomToTimeline()
        self.mouse_dragging = True  # Prevent mouseReleaseEvent from moving selection
        event.accept()

    def mousePressEvent(self, event):
        """Capture mouse press event"""
        event.accept()
        self.mouse_pressed = True
        self.mouse_dragging = False
        self.mouse_position = event.pos().x()
        self.scrollbar_position_previous = list(self.scrollbar_position)  # copy, don't alias

    def mouseReleaseEvent(self, event):
        """Capture mouse release event"""
        event.accept()

        # Handle the case where no dragging occurred (single click)
        if not self.mouse_dragging and not self.scroll_bar_rect.contains(event.pos()):
            # Center the scroll region at the click position (if outside the selection)
            click_pos = event.pos().x() / self.width()
            selection_width = self.scrollbar_position[1] - self.scrollbar_position[0]
            half_width = selection_width / 2

            # Calculate new left / right handles
            new_left_pos = click_pos - half_width
            new_right_pos = click_pos + half_width

            # If the new left position is less than 0, adjust both sides to fit within bounds
            if new_left_pos < 0.0:
                diff = -new_left_pos
                new_left_pos = 0.0
                new_right_pos = min(1.0, new_right_pos + diff)

            # If the new right position is greater than 1, adjust both sides to fit within bounds
            if new_right_pos > 1.0:
                diff = new_right_pos - 1.0
                new_right_pos = 1.0
                new_left_pos = max(0.0, new_left_pos - diff)

            # Update the scrollbar position to the newly calculated values
            self.scrollbar_position = [new_left_pos, new_right_pos, self.scrollbar_position[2], self.scrollbar_position[3]]

            # Trigger the resize and update
            self.delayed_resize_timer.start()
            self.update()

        # Finalize drag selection
        self.mouse_pressed = False
        self.mouse_dragging = False
        self.left_handle_dragging = False
        self.right_handle_dragging = False
        self.scroll_bar_dragging = False
        self.create_bar_dragging = False
        self.update()

    def set_handle_limits(self, left_handle, right_handle, is_left=False):
        """Set min/max limits on the bounds of the handles (to prevent invalid values)"""
        if left_handle < 0.0:
            left_handle = 0.0
            right_handle = self.scroll_bar_rect.width() / self.width()
        if right_handle > 1.0:
            left_handle = 1.0 - (self.scroll_bar_rect.width() / self.width())
            right_handle = 1.0

        # Don't allow handles to extend past each other
        diff = right_handle - left_handle

        # Adjust currently dragged handle (if exceeding min distance)
        if is_left and diff < self.min_distance:
            left_handle = right_handle - self.min_distance
        elif not is_left and diff < self.min_distance:
            right_handle = left_handle + self.min_distance

        return left_handle, right_handle

    def mouseMoveEvent(self, event):
        """Capture mouse events"""
        event.accept()

        # Get current mouse position
        mouse_pos = event.pos().x()
        if mouse_pos < 0:
            mouse_pos = 0
        elif mouse_pos > self.width():
            mouse_pos = self.width()

        # Set cursor (based on current mouse position)
        drag_threshold = 5
        if not self.mouse_dragging:
            if self.left_handle_rect.contains(event.pos()):
                self.setCursor(self.cursors.get('resize_x'))
            elif self.right_handle_rect.contains(event.pos()):
                self.setCursor(self.cursors.get('resize_x'))
            elif self.scroll_bar_rect.contains(event.pos()):
                self.setCursor(self.cursors.get('move'))
            else:
                self.setCursor(Qt.ArrowCursor)

        # Detect dragging (only if the user clicked and started dragging beyond the threshold)
        if self.mouse_pressed and not self.mouse_dragging:
            self.mouse_dragging = True
            if self.left_handle_rect.contains(event.pos()):
                self.left_handle_dragging = True
            elif self.right_handle_rect.contains(event.pos()):
                self.right_handle_dragging = True
            elif self.scroll_bar_rect.contains(event.pos()):
                self.scroll_bar_dragging = True
            elif abs(self.mouse_position - mouse_pos) > drag_threshold:
                # If clicking outside the current selection, initiate drag to create a new selection
                self.create_bar_dragging = True
            else:
                self.mouse_dragging = False

        # Handle dragging the selection (scroll bar dragging)
        if self.mouse_dragging:
            if self.left_handle_dragging:
                # Dragging the left handle to resize the selection
                delta = (self.mouse_position - mouse_pos) / self.width()
                new_left_pos = self.scrollbar_position_previous[0] - delta
                is_left = True

                if int(QCoreApplication.instance().keyboardModifiers() & Qt.ShiftModifier) > 0:
                    # SHIFT key pressed, move both handles
                    if (self.scrollbar_position_previous[1] + delta) - new_left_pos > self.min_distance:
                        new_right_pos = self.scrollbar_position_previous[1] + delta
                    else:
                        midpoint = (self.scrollbar_position_previous[1] + self.scrollbar_position_previous[0]) / 2
                        new_right_pos = midpoint + (self.min_distance / 2)
                        new_left_pos = midpoint - (self.min_distance / 2)
                else:
                    new_right_pos = self.scrollbar_position_previous[1]

                # Enforce limits (don't allow handles to go past each other, or out of bounds)
                new_left_pos, new_right_pos = self.set_handle_limits(new_left_pos, new_right_pos, is_left)

                self.scrollbar_position = [new_left_pos, new_right_pos, self.scrollbar_position[2], self.scrollbar_position[3]]
                self.delayed_resize_timer.start()

            elif self.right_handle_dragging:
                # Dragging the right handle to resize the selection
                delta = (self.mouse_position - mouse_pos) / self.width()
                new_right_pos = self.scrollbar_position_previous[1] - delta
                is_left = False

                if int(QCoreApplication.instance().keyboardModifiers() & Qt.ShiftModifier) > 0:
                    # SHIFT key pressed, move both handles
                    if new_right_pos - (self.scrollbar_position_previous[0] + delta) > self.min_distance:
                        new_left_pos = self.scrollbar_position_previous[0] + delta
                    else:
                        midpoint = (self.scrollbar_position_previous[1] + self.scrollbar_position_previous[0]) / 2
                        new_right_pos = midpoint + (self.min_distance / 2)
                        new_left_pos = midpoint - (self.min_distance / 2)
                else:
                    new_left_pos = self.scrollbar_position_previous[0]

                # Enforce limits
                new_left_pos, new_right_pos = self.set_handle_limits(new_left_pos, new_right_pos, is_left)

                self.scrollbar_position = [new_left_pos, new_right_pos, self.scrollbar_position[2], self.scrollbar_position[3]]
                self.delayed_resize_timer.start()

            elif self.scroll_bar_dragging:
                # Dragging the entire selection (scrolling the timeline)
                delta = (self.mouse_position - mouse_pos) / self.width()
                new_left_pos = self.scrollbar_position_previous[0] - delta
                new_right_pos = self.scrollbar_position_previous[1] - delta

                # Enforce limits
                new_left_pos, new_right_pos = self.set_handle_limits(new_left_pos, new_right_pos)

                self.scrollbar_position = [new_left_pos, new_right_pos, self.scrollbar_position[2], self.scrollbar_position[3]]
                get_app().window.TimelineScroll.emit(new_left_pos)

            elif self.create_bar_dragging:
                # Handle creating a new selection region
                new_pos = mouse_pos / self.width()

                if self.mouse_position < mouse_pos:
                    # Dragging to the right: set both handles to the starting position,
                    # then move the right handle (left handle stays where the drag started)
                    new_left_pos = self.mouse_position / self.width()
                    new_right_pos = new_pos
                else:
                    # Dragging to the left: set both handles to the starting position,
                    # then move the left handle (right handle stays where the drag started)
                    new_right_pos = self.mouse_position / self.width()
                    new_left_pos = new_pos

                # Enforce limits for the new selection
                new_left_pos, new_right_pos = self.set_handle_limits(new_left_pos, new_right_pos)
                self.scrollbar_position = [new_left_pos, new_right_pos, self.scrollbar_position[2], self.scrollbar_position[3]]
                self.delayed_resize_timer.start()

            # Force re-paint after any dragging
            self.update()

    def resizeEvent(self, event):
        """Widget resize event"""
        event.accept()
        self.delayed_size = self.size()
        self.delayed_resize_timer.start()

    def get_scroll_width(self):
        """Calculate the width of the scrollbar handle (i.e. selection width)"""
        # Get max width of timeline
        project_duration = get_app().project.get("duration")

        # Calculate scroll bar / selection width
        timeline_pixels_per_second = 100.0 / get_app().project.get("scale")
        timeline_project_width = project_duration * timeline_pixels_per_second
        scroll_ratio = self.scrollbar_position[3] / timeline_project_width
        scroll_width = scroll_ratio * self.width()
        scroll_width = min(scroll_width, self.width())
        return scroll_width, scroll_ratio

    def delayed_resize_callback(self):
        """Callback for resize event timer (to delay the resize event, and prevent lots of similar resize events)"""
        # Get max width of timeline
        project_duration = get_app().project.get("duration")
        normalized_scroll_width = self.scrollbar_position[1] - self.scrollbar_position[0]
        scroll_width_seconds = normalized_scroll_width * project_duration
        tick_pixels = 100
        if self.scrollbar_position[3] > 0.0:
            # Calculate the new zoom factor, based on pixels per tick
            zoom_factor = scroll_width_seconds / (self.scrollbar_position[3] / tick_pixels)

            # Set scroll width (and send signal)
            if zoom_factor > 0.0:
                self._syncing_backend = True
                try:
                    get_app().window.TimelineScroll.emit(self.scrollbar_position[0])
                    self.setZoomFactor(zoom_factor)
                finally:
                    self._syncing_backend = False

    # Capture wheel event to alter zoom/scale of widget
    def wheelEvent(self, event):
        event.accept()

        # Use async repaint to avoid recursive paint on some Qt/Windows paths.
        self.update()

    def setZoomFactor(self, zoom_factor, center=False, emit=True):
        """Set the current zoom factor (do not clamp width here — backend owns authoritative geometry)."""
        self.zoom_factor = zoom_factor
        if emit:
            get_app().window.TimelineZoom.emit(self.zoom_factor)
        if center:
            get_app().window.TimelineCenter.emit()

        # Force re-paint asynchronously
        self.update()

    def zoomIn(self):
        """Zoom into timeline"""
        if self.zoom_factor >= 10.0:
            new_factor = self.zoom_factor - 5.0
        elif self.zoom_factor >= 4.0:
            new_factor = self.zoom_factor - 2.0
        else:
            new_factor = self.zoom_factor * 0.8

        # Emit zoom signal
        self.setZoomFactor(new_factor, center=True)

    def zoomOut(self):
        """Zoom out of timeline"""
        if self.zoom_factor >= 10.0:
            new_factor = self.zoom_factor + 5.0
        elif self.zoom_factor >= 4.0:
            new_factor = self.zoom_factor + 2.0
        else:
            # Ensure zoom is reversable when using only keyboard zoom
            new_factor = min(self.zoom_factor * 1.25, 4.0)

        # Emit zoom signal
        self.setZoomFactor(new_factor, center=True)

    def update_scrollbars(self, new_positions):
        """Consume the current scroll bar positions from the webview timeline"""
        if self.mouse_dragging:
            return

        self.scrollbar_position = new_positions

        # Check for empty clips rects
        if not self.clip_rects:
            self.changed(None)

        # Disable auto center
        self.is_auto_center = False

        # Force re-paint asynchronously
        self.update()

    def handle_selection(self):
        # Force recalculation of clips and repaint
        self.changed(None)
        self.update()

    def timeline_resized(self):
        # Force recalculation of clips and repaint
        self.update()
        self.delayed_resize_timer.start()

    def update_playhead_pos(self, currentFrame):
        """Callback when position is changed"""
        self.current_frame = currentFrame

        # Schedule repaint (non-blocking)
        self.update()

    def handle_play(self):
        """Callback when play button is clicked"""
        self.is_auto_center = True

    def connect_playback(self):
        """Connect playback signals"""
        self.win.preview_thread.position_changed.connect(self.update_playhead_pos)
        self.win.PlaySignal.connect(self.handle_play)

    def ignore_updates_callback(self, ignore, show_wait=True):
        """Ignore updates callback - used to stop updating this widget during batch updates"""
        if not ignore and self.ignore_updates:
            # Force recalculation and repaint
            self.ignore_updates = ignore
            self.changed(None)
            self.update()
        self.ignore_updates = ignore

    def __init__(self, *args):
        # Invoke parent init
        QWidget.__init__(self, *args)

        # Translate object
        _ = get_app()._tr

        # Init default values
        self.leftHandle = None
        self.rightHandle = None
        self.centerHandle = None
        self.mouse_pressed = False
        self.mouse_dragging = False
        self.mouse_position = None
        self.zoom_factor = 15.0
        self.scrollbar_position = [0.0, 0.0, 0.0, 0.0]
        self.scrollbar_position_previous = [0.0, 0.0, 0.0, 0.0]
        self.scrollbar_zoom_previous = [0.0, 0.2, 0.0, 0.0]
        self.left_handle_rect = QRectF()
        self.left_handle_dragging = False
        self.right_handle_rect = QRectF()
        self.right_handle_dragging = False
        self.scroll_bar_rect = QRectF()
        self.scroll_bar_dragging = False
        self.clip_rects = []
        self.clip_rects_selected = []
        self.marker_rects = []
        self.current_frame = 0
        self.is_auto_center = True
        self.min_distance = 0.002
        self.ignore_updates = False
        self._syncing_backend = False

        # Load icon (using display DPI)
        self.cursors = {}
        for cursor_name in ["move", "resize_x", "hand"]:
            icon = QIcon(":/cursors/cursor_%s.png" % cursor_name)
            self.cursors[cursor_name] = QCursor(icon.pixmap(24, 24))

        # Init Qt widget's properties (background repainting, etc...)
        super().setAttribute(Qt.WA_OpaquePaintEvent)
        super().setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # Add self as listener to project data updates (used to update the timeline)
        get_app().updates.add_listener(self)

        # Set mouse tracking
        self.setMouseTracking(True)

        # Get a reference to the window object
        self.win = get_app().window

        # Connect zoom functionality
        self.win.TimelineScrolled.connect(self.update_scrollbars)
        self.win.TimelineResize.connect(self.timeline_resized)
        self.win.IgnoreUpdates.connect(self.ignore_updates_callback)
        self.win.TimelineZoom.connect(lambda z: self.setZoomFactor(z, emit=False))

        # Connect Selection signals
        self.win.SelectionChanged.connect(self.handle_selection)

        # Show Property timer
        # Timer to use a delay before sending MaxSizeChanged signals (so we don't spam libopenshot)
        self.delayed_size = None
        self.delayed_resize_timer = QTimer(self)
        self.delayed_resize_timer.setInterval(100)
        self.delayed_resize_timer.setSingleShot(True)
        self.delayed_resize_timer.timeout.connect(self.delayed_resize_callback)



#!/usr/bin/python3
"""
 @file
 @brief Display all available string translations for each translation file
 @author Jonathan Thomas <jonathan@openshot.org>
 @author Frank Dana <ferdnyc AT gmail com>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import re
import fnmatch
import sys
from PyQt5.QtCore import QLocale, QLibraryInfo, QTranslator, QCoreApplication


# Get the absolute path of this project
language_path = os.path.dirname(os.path.abspath(__file__))

# Get app instance
app = QCoreApplication(sys.argv)

# Load POT template (all English strings)
all_templates = ['OpenShot.pot', 'OpenShot_transitions.pot', 'OpenShot_blender.pot']
for template_name in all_templates:
    POT_source = open(os.path.join(language_path, 'OpenShot', template_name)).read()
    all_strings = re.findall('^msgid \"(.*)\"', POT_source, re.MULTILINE)

    print("Scanning {} strings in all translation files...".format(len(all_strings)))

    # Loop through folders/languages
    for filename in fnmatch.filter(os.listdir(language_path), 'OpenShot*.qm'):
        lang_code = filename[:-3]
        # Install language
        translator = QTranslator(app)

        # Load translation
        if translator.load(lang_code, language_path):
            app.installTranslator(translator)

            print("\n=================================================")
            print("Showing translations for {}".format(filename))
            print("=================================================")
            # Loop through all test strings
            for source_string in all_strings:
                translated_string = app.translate("", source_string)
                if source_string != translated_string:
                    print('  {} => {}'.format(source_string,translated_string))

                if "%s" in source_string or "%s(" in source_string or "%d" in source_string:
                    if source_string.count('%') != translated_string.count('%'):
                        raise(Exception('Invalid string replacement found: "%s" vs "%s" [%s]' %
                              (translated_string, source_string, lang_code)))

            # Remove translator
            app.removeTranslator(translator)



"""
 @file
 @brief This file is for legacy support of OpenShot 1.x project files
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import uuid


class OpenShotFile:
    """The generic file object for OpenShot"""

    # ----------------------------------------------------------------------
    def __init__(self, project=None):
        """Constructor"""
        self.project = project

        # init the variables for the File Object
        self.name = ""  # short / friendly name of the file
        self.length = 0.0  # length in seconds
        self.videorate = (30, 0)  # audio rate or video framerate
        self.file_type = ""  # video, audio, image, image sequence
        self.max_frames = 0.0
        self.fps = 0.0
        self.height = 0
        self.width = 0
        self.label = ""  # user description of the file
        self.thumb_location = ""  # file uri of preview thumbnail
        self.ttl = 1  # time-to-live - only used for image sequence.  Represents the # of frames per image.

        self.unique_id = str(uuid.uuid1())
        self.parent = None
        self.project = project  # reference to project

        self.video_codec = ""
        self.audio_codec = ""
        self.audio_frequency = ""
        self.audio_channels = ""


class OpenShotFolder:
    """The generic folder object for OpenShot"""

    # ----------------------------------------------------------------------
    def __init__(self, project=None):
        """Constructor"""

        # Init the variables for the Folder Object
        self.name = ""  # short / friendly name of the folder
        self.location = ""  # file system location
        self.parent = None
        self.project = project

        self.label = ""  # user description of the folder
        self.unique_id = str(uuid.uuid1())

        # init the list of files & folders
        # this list can contain OpenShotFolder or OpenShotFile objects
        # the order of this list determines the order of the tree items
        self.items = []

        # this queue holds files that are currently being added. this prevents
        # duplicate files to be added at the same time
        self.queue = []



"""
 @file
 @brief This file contains a theme's colors and UI dimensions
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2024 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from ..base import BaseTheme


class HumanityDarkTheme(BaseTheme):
    def __init__(self, app):
        super().__init__(app)
        self.style_sheet = """
QToolTip {
    color: #ffffff;
    background-color: #2a82da;
    border: 0px solid white;
}

QComboBox::item {
    height: 24px;
}

QComboBox {
    combobox-popup: 0;
}

.property_value {
    foreground-color: #217dd4;
    background-color: #565656;
}

.zoom_slider_playhead {
    background-color: #ff0024;
}

QWidget#videoPreview {
    background-color: #191919;
}

QLabel#lblMissingFileHint,
QLabel#lblMissingFilePath {
    color: #b8b8b8;
}
        """

    def apply_theme(self):
        super().apply_theme()

        from classes import ui_util
        from classes.logger import log
        from PyQt5.QtWidgets import QStyleFactory

        log.info("Setting Fusion dark palette")
        self.app.setStyle(QStyleFactory.create("Fusion"))
        dark_palette = ui_util.make_dark_palette(self.app.palette())
        self.app.setPalette(dark_palette)
        self.app.setStyleSheet(self.compose_stylesheet())

        # Apply timeline theme
        self.app.window.timeline.apply_theme("""
            .keyframe-panel-row {
              background: #2f2f2f;
              border: 0px solid transparent;
            }
            .keyframe-panel-curve {
              color: #4b92ad;
            }
            .keyframe-panel-point {
              background: #4b92ad;
              border: 1px solid #7dc3dd;
              opacity: 0.70;
            }
        """)

        # Emit signal
        self.app.window.ThemeChangedSignal.emit(self)

class Retro(BaseTheme):
    def __init__(self, app):
        super().__init__(app)
        self.style_sheet = """
QComboBox::item {
    height: 24px;
}

QMainWindow::separator:hover {
    background: #dedede;
}

.property_value {
    foreground-color: #217dd4;
    background-color: #7f7f7f;
}

.zoom_slider_playhead {
    background-color: #ff0024;
}

QWidget#videoPreview {
    background-color: #dedede;
}

QLabel#lblMissingFileHint,
QLabel#lblMissingFilePath {
    color: #5a5a5a;
}

QComboBox {
    combobox-popup: 0;
}
        """

    def apply_theme(self):
        super().apply_theme()

        # Apply timeline theme
        self.app.window.timeline.apply_theme("""
            body {
              background: #f0f0f0;
            }
            #ruler_time {
              color: #c9c9c9;
            }
           .ruler_time {
              color: #c9c9c9;
            }
            #ruler_label {
              height: 43px;
              background: linear-gradient(to bottom, #3c3c3c, #0a070a);
              margin-bottom: 0px;
            }
            #scrolling_ruler {
              background: linear-gradient(to bottom, #3c3c3c, #0a070a);
              margin-bottom: 0px;
            }
            .track_name {
              margin-top: 8px;
              color: #000000;
              background: linear-gradient(to bottom, #dedddd, #d2d2d3);
              box-shadow: none;
            }
            .track {
              margin-top: 8px;
              background: #e5e7ea;
              box-shadow: none;
            }
            .track-resize-handle {
              background-color: #BEBFC1;
            }
            .track-resize-handle:hover {
              background-color: #F7F8FA;
            }
            .transition_top {
              background: none;
              border-radius: 0px;
            }
            .transition {
              border: 1px solid #0192c1;
              border-radius: 0px;
              box-shadow: none;
            }
            .clip {
              border-radius: 0px;
              background: #fedc66;
              border: 1px solid #cd8d00;
              box-shadow: none;
            }
            .ui-selected {
                filter: brightness(1.1);
            }
            .clip_label {
              color: #383730;
            }
            .clip_effects {
              background: rgba(54, 25, 25, 0.6);
            }
            .point_bezier {
              background-image: url(../themes/humanity/images/keyframe-bezier.svg);
            }
            .point_linear {
              background-image: url(../themes/humanity/images/keyframe-linear.svg);
            }
            .point_constant {
              background-image: url(../themes/humanity/images/keyframe-constant.svg);
            }
            .track-keyframe-panel-disabled {
              background-image: url(../themes/humanity/images/retro-track-keyframe-panel-show-disabled.svg);
            }
            .track-keyframe-panel-enabled {
              background-image: url(../themes/humanity/images/retro-track-keyframe-panel-show-enabled.svg);
            }
            .track-add-above-disabled {
              background-image: url(../themes/humanity/images/track-add-above-disabled.svg);
            }
            .track-add-above-enabled {
              background-image: url(../themes/humanity/images/track-add-above-enabled.svg);
            }
            .track-add-below-disabled {
              background-image: url(../themes/humanity/images/track-add-below-disabled.svg);
            }
            .track-add-below-enabled {
              background-image: url(../themes/humanity/images/track-add-below-enabled.svg);
            }
            .track-delete-disabled {
              background-image: url(../themes/humanity/images/track-delete-disabled.svg);
            }
            .track-delete-enabled {
              background-image: url(../themes/humanity/images/track-delete-enabled.svg);
            }
            .track-locked-disabled {
              background-image: url(../themes/humanity/images/retro-track-locked-disabled.svg);
            }
            .track-locked-enabled {
              background-image: url(../themes/humanity/images/retro-track-locked-enabled.svg);
            }
            .track-unlocked-disabled {
              background-image: url(../themes/humanity/images/retro-track-unlocked-disabled.svg);
            }
            .track-unlocked-enabled {
              background-image: url(../themes/humanity/images/retro-track-unlocked-enabled.svg);
            }
            .keyframe-panel-add {
              background-image: url(../themes/humanity/images/keyframe-panel-add.svg);
            }
            .keyframe-panel-row {
              background: #e5e7ea;
              border: 0px solid transparent;
            }
            .keyframe-panel-curve {
              color: #4b92ad;
            }
            .keyframe-panel-point {
              background: #4b92ad;
              border: 1px solid #3a748a;
              opacity: 0.72;
            }
        """)

        # Emit signal
        self.app.window.ThemeChangedSignal.emit(self)



"""
 @file
 @brief Painter for cached playback segments.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
"""

from PyQt5.QtCore import QRectF, Qt
from PyQt5.QtGui import QColor, QPainter

from .base import BasePainter


class PlaybackCachePainter(BasePainter):
    """Render cached playback ranges as a thin bar under the ruler."""

    def update_theme(self):
        color = getattr(self.w.theme, "playback_cache_color", QColor("#4B92AD"))
        if not isinstance(color, QColor) or not color.isValid():
            color = QColor("#4B92AD")
        self.cache_color = color

        height = getattr(self.w.theme, "playback_cache_height", 5.0)
        try:
            height = float(height)
        except (TypeError, ValueError):
            height = 5.0
        if height <= 0.0:
            height = 5.0
        self.cache_height = height

    def paint(self, painter: QPainter):
        ranges = getattr(self.w, "_playback_cache_ranges", None)
        if not ranges:
            return

        pps = float(getattr(self.w, "pixels_per_second", 0.0) or 0.0)
        if pps <= 0.0:
            return

        area = QRectF(
            self.w.track_name_width,
            self.w.ruler_height,
            self.w.width() - self.w.track_name_width - self.w.scroll_bar_thickness,
            self.w.height() - self.w.ruler_height - self.w.scroll_bar_thickness,
        )
        if area.width() <= 0.0 or area.height() <= 0.0:
            return

        bar_height = min(self.cache_height, area.height())
        if bar_height <= 0.0:
            return

        offset_px = float(getattr(self.w, "h_scroll_offset", 0.0) or 0.0)

        painter.save()
        painter.setClipRect(area)
        painter.setPen(Qt.NoPen)
        painter.setBrush(self.cache_color)

        top = area.top()
        for start_seconds, end_seconds in ranges:
            if end_seconds <= start_seconds:
                continue
            start_px = self.w.track_name_width + start_seconds * pps - offset_px
            end_px = self.w.track_name_width + end_seconds * pps - offset_px
            width = end_px - start_px
            if width <= 0.5:
                continue
            rect = QRectF(start_px, top, width, bar_height)
            rect = rect.intersected(area)
            if rect.isNull():
                continue
            painter.fillRect(rect, self.cache_color)

        painter.restore()



"""
 @file
 @brief This file contains the ThemeManager singleton, used to easily switch UI themes
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2024 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import threading
from enum import Enum


class ThemeName(Enum):
    """Friendly UI theme names used in settings"""
    RETRO = "Retro"
    HUMANITY_DARK = "Humanity: Dark"
    COSMIC = "Cosmic Dusk"

    @staticmethod
    def get_sorted_theme_names():
        """Return a sorted list of theme names"""
        return sorted([theme.value for theme in ThemeName])

    @staticmethod
    def find_by_name(name):
        """Return a theme ENUM which matches a name"""
        for theme in ThemeName:
            if theme.value == name:
                return theme
        return ThemeName.HUMANITY_DARK


class ThemeManager:
    """Singleton Theme Manager class, used to easily switch between UI themes"""
    _instance = None
    _lock = threading.Lock()

    def __new__(cls, app=None):
        """Override new method, so the same instance is always returned (i.e. singleton)"""
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super(ThemeManager, cls).__new__(cls)
                    cls._instance.app = app
                    cls._instance.original_style = app.style().objectName() if app else None
                    cls._instance.original_palette = app.palette() if app else None
                    cls._instance.current_theme = None
        return cls._instance

    def apply_theme(self, name):
        """Apply a new UI theme. Expects a ThemeName ENUM as the arg."""
        theme_enum = ThemeName.find_by_name(name)

        if theme_enum == ThemeName.HUMANITY_DARK:
            from themes.humanity.theme import HumanityDarkTheme
            self.current_theme = HumanityDarkTheme(self.app)
        elif theme_enum == ThemeName.RETRO:
            from themes.humanity.theme import Retro
            self.current_theme = Retro(self.app)
        elif theme_enum == ThemeName.COSMIC:
            from themes.cosmic.theme import CosmicTheme
            self.current_theme = CosmicTheme(self.app)

        # Set name on theme instance
        self.current_theme.name = theme_enum.value

        # Apply theme
        self.current_theme.apply_theme()
        return self.current_theme

    def get_current_theme(self):
        """Return the current theme instance, or None if no theme is applied."""
        return self.current_theme



"""
 @file
 @brief This file contains unit tests for the Query class
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import sys
import os
import json

import unittest

import openshot

from PyQt5.QtGui import QGuiApplication

# Import parent folder (so it can find other imports)
PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from classes.app import OpenShotApp
from classes.query import Clip, File, Transition
from classes import info
from classes.project_data import ProjectDataStore
from classes.updates import UpdateManager
from qt_test_app import ensure_app_state as ensure_qt_app_state

info.LOG_LEVEL_CONSOLE = "ERROR"


class DummySettings:
    def __init__(self):
        self.values = {
            "default-profile": "HD 720p 30 fps",
            "default-samplerate": 48000,
            "default-channels": 2,
            "legacy-based-timeline": False,
        }

    def get(self, key):
        return self.values.get(key)

    def set(self, key, value):
        self.values[key] = value


def ensure_open_shot_app():
    """Ensure an OpenShot QApplication exists before tests run."""

    instance = QGuiApplication.instance()
    if instance:
        return ensure_qt_app_state(
            instance,
            DummySettings,
            project_factory=ProjectDataStore,
            updates_factory=UpdateManager,
            extra_attrs={"window": None},
        ), False

    app = OpenShotApp(sys.argv, mode="unittest")
    return ensure_qt_app_state(app, DummySettings), True


class QueryTests(unittest.TestCase):
    """ Unit test class for Query class """

    @classmethod
    def setUpClass(cls):
        """ Init unit test data """
        # Create or reuse Qt application
        cls.app, cls._owns_app = ensure_open_shot_app()
        cls.clip_ids = []
        cls.file_ids = []
        cls.transition_ids = []

        clips = []

        # Insert some clips into the project data
        for num in range(5):
            # Create clip
            c = openshot.Clip(os.path.join(info.IMAGES_PATH, "AboutLogo.png"))
            c.Position(num * 10.0)
            c.End(5.0)

            # Parse JSON
            clip_data = json.loads(c.Json())

            # Insert into project data
            query_clip = Clip()
            query_clip.data = clip_data
            query_clip.save()

            # Keep track of the ids
            cls.clip_ids.append(query_clip.id)
            clips.append(query_clip)

        # Insert some files into the project data
        for num in range(5):
            # Create file
            r = openshot.DummyReader(openshot.Fraction(24, 1), 640, 480, 44100, 2, 30.0)

            # Parse JSON
            file_data = json.loads(r.Json())

            # Insert into project data
            query_file = File()
            query_file.data = file_data
            query_file.data["path"] = os.path.join(info.IMAGES_PATH, "AboutLogo.png")
            query_file.data["media_type"] = "image"
            query_file.save()

            # Keep track of the ids
            cls.file_ids.append(query_file.id)

        # Insert some transitions into the project data
        for c in clips:
            # Create mask object
            t = openshot.Mask()
            # Place over the last second of current clip
            pos = c.data.get("position", 0.0)
            start = c.data.get("start", 0.0)
            end = c.data.get("end", 0.0)
            t.Position((pos - start + end) - 1.0)
            t.End(1.0)

            # Insert into project data
            transitions_data = json.loads(t.Json())
            query_transition = Transition()
            query_transition.data = transitions_data
            query_transition.save()

            # Keep track of the ids
            cls.transition_ids.append(query_transition.id)

        # Don't keep the full query objects around
        del clips

    @classmethod
    def tearDownClass(cls):
        """ Clean up after running all tests in the class. """
        if getattr(cls, "_owns_app", False) and cls.app:
            cls.app.quit()

    def test_add_clip(self):

        # Find number of clips in project
        num_clips = len(Clip.filter())

        # Create clip
        c = openshot.Clip(os.path.join(info.IMAGES_PATH, "AboutLogo.png"))
        clip_data = json.loads(c.Json())

        # Insert into project data
        query_clip = Clip()
        query_clip.data = clip_data
        query_clip.save()

        self.assertTrue(query_clip)
        self.assertEqual(len(Clip.filter()), num_clips + 1)

        # Save the clip again (which should not change the total # of clips)
        query_clip.save()
        self.assertEqual(len(Clip.filter()), num_clips + 1)

    def test_update_clip(self):
        """ Test the Clip.save method """

        update_id = self.clip_ids[0]
        clip = Clip.get(id=update_id)
        self.assertTrue(clip)

        # Update clip
        clip.data["layer"] = 2
        clip.data["title"] = "My Title"
        clip.save()

        # Verify updated data
        clip = Clip.get(id=update_id)
        self.assertEqual(clip.data["layer"], 2)
        self.assertEqual(clip.data["title"], "My Title")

        clips = Clip.filter(layer=2)
        self.assertEqual(len(clips), 1)

    def test_delete_clip(self):
        """ Test the Clip.delete method """

        delete_id = self.clip_ids[4]
        clip = Clip.get(id=delete_id)
        self.assertTrue(clip)

        clip.delete()

        # Verify deleted data
        deleted_clip = Clip.get(id=delete_id)
        self.assertFalse(deleted_clip)

        # Delete clip again (should do nothing)
        clip.delete()
        deleted_clip = Clip.get(id=delete_id)
        self.assertFalse(deleted_clip)

    def test_filter_clip(self):
        """ Test the Clip.filter method """

        clips = Clip.filter(id=self.clip_ids[0])
        self.assertTrue(clips)

        # Do not find a clip
        clips = Clip.filter(id="invalidID")
        self.assertEqual(len(clips), 0)

    def test_get_clip(self):
        """ Test the Clip.get method """

        clip = Clip.get(id=self.clip_ids[1])
        self.assertTrue(clip)

        # Do not find a clip
        clip = Clip.get(id="invalidID")
        self.assertEqual(clip, None)

    def test_intersect(self):
        """ Test special filter argument 'intersect' """

        trans = Transition.get(id=self.transition_ids[0])
        self.assertTrue(trans)

        pos = trans.data.get("position", -1.0)
        duration = trans.data.get("duration", -1.0)
        self.assertTrue(pos >= 0.0)
        self.assertTrue(duration >= 0.0)

        time = pos + (duration / 2)

        def get_times(item):
            pos = item.data.get("position", -1.0)
            end = pos + item.data.get("duration", -1.0)
            return (pos, end)

        t_intersect = Transition.filter(intersect=time)
        t_ids = [t.id for t in t_intersect]
        t_all = Transition.filter()
        t_rest = [x for x in t_all if x.id not in t_ids]

        c_intersect = Clip.filter(intersect=time)
        c_ids = [c.id for c in c_intersect]
        c_all = Clip.filter()
        c_rest = [x for x in c_all if x.id not in c_ids]

        for item in t_intersect + c_intersect:
            item_id = item.id
            pos, end = get_times(item)
            self.assertTrue(pos <= time)
            self.assertTrue(time <= end)
        for item in t_rest + c_rest:
            item_id = item.id
            pos, end = get_times(item)
            if pos < time:
                self.assertTrue(end <= time)
            if end > time:
                self.assertTrue(pos >= time)

    def test_update_File(self):
        """ Test the File.save method """

        update_id = self.file_ids[0]
        file = File.get(id=update_id)
        self.assertTrue(file)

        # Update File
        file.data["height"] = 1080
        file.data["width"] = 1920
        file.save()

        # Verify updated data
        file = File.get(id=update_id)
        self.assertEqual(file.data["height"], 1080)
        self.assertEqual(file.data["width"], 1920)

    def test_delete_File(self):
        """ Test the File.delete method """

        delete_id = self.file_ids[4]
        file = File.get(id=delete_id)
        self.assertTrue(file)

        file.delete()

        # Verify deleted data
        deleted_file = File.get(id=delete_id)
        self.assertFalse(deleted_file)

        # Delete File again (should do nothing)
        file.delete()
        deleted_file = File.get(id=delete_id)
        self.assertFalse(deleted_file)

    def test_filter_File(self):
        """ Test the File.filter method """

        files = File.filter(id=self.file_ids[0])
        self.assertTrue(files)

        # Do not find a File
        files = File.filter(id="invalidID")
        self.assertEqual(len(files), 0)

    def test_get_File(self):
        """ Test the File.get method """

        file = File.get(id=self.file_ids[1])
        self.assertTrue(file)

        # Do not find a File
        file = File.get(id="invalidID")
        self.assertEqual(file, None)

    def test_get_File_by_path_with_alternate_separators(self):
        """File.get(path=...) should ignore slash direction differences."""
        file_obj = File.get(id=self.file_ids[0])
        self.assertTrue(file_obj)

        original_path = file_obj.data.get("path")
        self.assertTrue(original_path)

        if "\\" in original_path:
            alt_path = original_path.replace("\\", "/")
        else:
            alt_path = original_path.replace("/", "\\")

        matched = File.get(path=alt_path)
        self.assertTrue(matched)
        self.assertEqual(matched.id, file_obj.id)

    def test_add_file(self):

        # Find number of files in project
        num_files = len(File.filter())

        # Create file
        r = openshot.DummyReader(openshot.Fraction(24, 1), 640, 480, 44100, 2, 30.0)
        file_data = json.loads(r.Json())

        # Insert into project data
        query_file = File()
        query_file.data = file_data
        query_file.data["path"] = os.path.join(info.IMAGES_PATH, "AboutLogo.png")
        query_file.data["media_type"] = "image"

        query_file.save()
        self.assertTrue(query_file)
        self.assertEqual(len(File.filter()), num_files + 1)

        # Save the file again (which should not change the total # of files)
        query_file.save()
        self.assertEqual(len(File.filter()), num_files + 1)



"""
 @file
 @brief ComfyUI workflow template discovery and classification helpers.
"""

import copy
import json
import os

from classes import info
from classes.logger import log


IMAGE_INPUT_TYPES = {
    "loadimage",
    "load image",
}
VIDEO_INPUT_TYPES = {
    "loadvideo",
    "load video",
    "vhs_loadvideo",
}
AUDIO_INPUT_TYPES = {
    "loadaudio",
    "load audio",
}

IMAGE_OUTPUT_TYPES = {
    "saveimage",
    "save image",
}
VIDEO_OUTPUT_TYPES = {
    "savevideo",
    "save video",
}
AUDIO_OUTPUT_TYPES = {
    "saveaudio",
    "save audio",
}

KNOWN_NODE_TYPES = {
    # Input
    "checkpointloadersimple",
    "unetloader",
    "cliptextencode",
    "cliploader",
    "vaeloader",
    "loadimage",
    "loadvideo",
    "vhs_loadvideo",
    "loadaudio",
    # Core built-in/OpenShot workflows
    "vaeencode",
    "vaedecode",
    "ksampler",
    "upscalemodelloader",
    "imageupscalewithmodel",
    "videoslice",
    "video slice",
    "getvideocomponents",
    "createvideo",
    "saveimage",
    "savevideo",
    "saveaudio",
    "getimagesize",
    "vhs_getimagecount",
    "save srt",
    "emptylatentimage",
    "emptyhunyuanlatentvideo",
    "wan22imagetovideolatent",
    "wanimagetoimageapi",
    "wantexttoimageapi",
    "imageonlycheckpointloader",
    "modelsamplingsd3",
    "svd_img2vid_conditioning",
    "videolinearcfgguidance",
    "emptylatentaudio",
    "vaedecodeaudio",
    "previewany",
    "apply whisper",
    "riff vfi",
    "rife vfi",
    "downloadandloadtransnetmodel",
    "transnetv2_run",
    "selectvideo",
    "stableaudioprojectionmodel",
    "stableaudiomodelloader",
    "stableaudioemptylatentaudio",
    "stableaudioembedding",
    "kdiffusionsampler",
    "stableaudiovaedecode",
    "videocombine",
    "imagescaleby",
    "imagetosimage",
    "imagetoimage",
    "imagescaleto",
    "imageblur",
    "imagecompositemasked",
    "imageblend",
    "controlnetloader",
    "controlnetapply",
    "controlnetapplyadvanced",
    "canny",
    "depthanythingpreprocessor",
    "depthanythingv2preprocessor",
    "midas-depthmappreprocessor",
    "zoe-depthmappreprocessor",
    "zoe_depthanythingpreprocessor",
    # Video Helper Suite
    "vhs_batchmanager",
    "vhs_loadvideo",
    "vhs_loadvideopath",
    "vhs_loadvideoffmpegpath",
    "vhs_videocombine",
    "vhs_videoinfo",
    "vhs_videoinfoloaded",
    "vhs_videoinfosource",
    # ComfyUI-segment-anything-2
    "downloadandloadsam2model",
    "sam2segmentation",
    "sam2autosegmentation",
    "sam2videosegmentationaddpoints",
    "sam2videosegmentation",
    # OpenShot-ComfyUI (custom SAM2)
    "openshotdownloadandloadsam2model",
    "openshotsam2segmentation",
    "openshotsam2videosegmentationaddpoints",
    "openshotsam2videosegmentationchunked",
    "openshotimageblurmasked",
    "openshotimagehighlightmasked",
}


class ComfyTemplateRegistry:
    """Discovers ComfyUI templates from built-in + user folders."""

    def __init__(self):
        self._cache = None
        self._cache_signature = None

    @staticmethod
    def _is_ignored_filename(name):
        return str(name or "").strip().lower() in ("debug.json", "debug_error.json", "debug_sent.json")

    def _template_roots(self):
        return [
            (os.path.join(info.PATH, "comfyui"), False),
            (info.COMFYUI_PATH, True),
        ]

    def _current_signature(self):
        signature = []
        for folder, _is_user in self._template_roots():
            if not os.path.isdir(folder):
                continue
            for name in sorted(os.listdir(folder)):
                if not name.lower().endswith(".json"):
                    continue
                if self._is_ignored_filename(name):
                    continue
                path = os.path.join(folder, name)
                try:
                    stat = os.stat(path)
                    signature.append((path, stat.st_mtime_ns, stat.st_size))
                except OSError:
                    continue
        return tuple(signature)

    def discover(self, force=False):
        signature = self._current_signature()
        if not force and self._cache is not None and signature == self._cache_signature:
            return self._cache

        templates = []
        existing_ids = set()
        for folder, is_user in self._template_roots():
            if not os.path.isdir(folder):
                continue
            for name in sorted(os.listdir(folder)):
                if not name.lower().endswith(".json"):
                    continue
                if self._is_ignored_filename(name):
                    continue
                path = os.path.join(folder, name)
                template = self._load_template(path, is_user=is_user, existing_ids=existing_ids)
                if template is None:
                    continue
                templates.append(template)

        templates.sort(key=lambda t: (int(t.get("sort_order", 99999)), str(t.get("display_name", "")).lower()))
        self._cache = templates
        self._cache_signature = signature
        return templates

    def _load_template(self, path, is_user, existing_ids):
        try:
            with open(path, "r", encoding="utf-8") as handle:
                payload = json.load(handle)
        except Exception as ex:
            log.warning("Skipping invalid ComfyUI template JSON %s: %s", path, ex)
            return None
        if not isinstance(payload, dict):
            log.warning("Skipping invalid ComfyUI template JSON %s: root must be an object", path)
            return None

        workflow = self._extract_workflow(payload)
        if workflow is None:
            log.warning("Skipping invalid ComfyUI template JSON %s: no valid workflow graph found", path)
            return None

        node_types = []
        input_types = set()
        output_types = set()
        unknown_node_types = set()
        needs_prompt = False
        for node in workflow.values():
            if not isinstance(node, dict):
                continue
            class_type = str(node.get("class_type", "")).strip()
            if not class_type:
                continue
            inputs = node.get("inputs", {})
            if not isinstance(inputs, dict):
                inputs = {}
            class_key = class_type.lower().replace("_", "")
            class_flat = class_type.lower().strip()
            node_types.append(class_type)

            text_value = inputs.get("text", None)
            if isinstance(text_value, str):
                meta = node.get("_meta", {})
                meta_title = ""
                if isinstance(meta, dict):
                    meta_title = str(meta.get("title", "")).strip().lower()
                if "textencode" in class_key or "prompt" in meta_title:
                    needs_prompt = True

            if class_flat in IMAGE_INPUT_TYPES or class_key in IMAGE_INPUT_TYPES:
                input_types.add("image")
            if class_flat in VIDEO_INPUT_TYPES or class_key in VIDEO_INPUT_TYPES:
                input_types.add("video")
            if class_flat in AUDIO_INPUT_TYPES or class_key in AUDIO_INPUT_TYPES:
                input_types.add("audio")

            if class_flat in IMAGE_OUTPUT_TYPES or class_key in IMAGE_OUTPUT_TYPES:
                output_types.add("image")
            if class_flat in VIDEO_OUTPUT_TYPES or class_key in VIDEO_OUTPUT_TYPES:
                output_types.add("video")
            if class_flat in AUDIO_OUTPUT_TYPES or class_key in AUDIO_OUTPUT_TYPES:
                output_types.add("audio")

            if class_flat not in KNOWN_NODE_TYPES and class_key not in KNOWN_NODE_TYPES:
                unknown_node_types.add(class_type)

        if unknown_node_types:
            log.warning(
                "ComfyUI template has unknown node types (%s): %s",
                os.path.basename(path),
                ", ".join(sorted(unknown_node_types)),
            )

        override_category = str(payload.get("menu_category") or payload.get("category") or "").strip().lower()
        override_menu_parent = str(payload.get("menu_parent") or "").strip()
        override_input_type = str(payload.get("input_type") or payload.get("source_type") or "").strip().lower()
        override_output_type = str(payload.get("output_type") or payload.get("media_output") or "").strip().lower()
        override_icon = str(payload.get("action_icon") or payload.get("icon") or "").strip()
        override_open_dialog = payload.get("open_dialog", None)
        needs_reference_image = bool(payload.get("needs_reference_image", False))

        inferred_category = "unknown"
        requires_source = bool(input_types)
        if output_types:
            inferred_category = "enhance" if requires_source else "create"
        else:
            if override_category not in ("create", "enhance", "unknown"):
                log.warning(
                    "ComfyUI template category unknown (%s): no output nodes detected",
                    os.path.basename(path),
                )
        if override_category in ("create", "enhance", "unknown"):
            inferred_category = override_category
        if override_input_type in ("image", "video", "audio"):
            input_types = {override_input_type}

        template_id = str(payload.get("template_id") or payload.get("id") or "").strip()
        if not template_id:
            template_id = os.path.splitext(os.path.basename(path))[0]

        unique_id = template_id
        suffix = 2
        while unique_id in existing_ids:
            unique_id = "{}__{}".format(template_id, suffix)
            suffix += 1
        existing_ids.add(unique_id)

        display_name = self._extract_name(payload, path)
        if is_user:
            display_name = "(User) {}".format(display_name)

        try:
            sort_order = int(payload.get("menu_order", 99999))
        except (TypeError, ValueError):
            sort_order = 99999
        primary_output = self._primary_output_type(output_types)
        if override_output_type in ("image", "video", "audio", "unknown"):
            primary_output = override_output_type

        open_dialog = None
        if isinstance(override_open_dialog, bool):
            open_dialog = override_open_dialog

        return {
            "id": unique_id,
            "template_id": template_id,
            "display_name": display_name,
            "path": path,
            "is_user": is_user,
            "category": inferred_category,
            "input_types": sorted(input_types),
            "output_types": sorted(output_types),
            "primary_output": primary_output,
            "sort_order": sort_order,
            "workflow": workflow,
            "node_types": node_types,
            "needs_prompt": needs_prompt,
            "action_icon": override_icon,
            "open_dialog": open_dialog,
            "menu_parent": override_menu_parent,
            "needs_reference_image": needs_reference_image,
        }

    def _primary_output_type(self, output_types):
        if "video" in output_types:
            return "video"
        if "image" in output_types:
            return "image"
        if "audio" in output_types:
            return "audio"
        return "unknown"

    def _extract_name(self, payload, path):
        fields = [
            payload.get("name"),
            payload.get("title"),
            payload.get("workflow_name"),
        ]
        metadata = payload.get("metadata")
        if isinstance(metadata, dict):
            fields.extend([metadata.get("name"), metadata.get("title")])

        for value in fields:
            text = str(value or "").strip()
            if text:
                return text

        return os.path.splitext(os.path.basename(path))[0]

    def _extract_workflow(self, payload):
        if self._looks_like_workflow(payload):
            return payload
        if isinstance(payload, dict):
            workflow = payload.get("workflow")
            if self._looks_like_workflow(workflow):
                return workflow
        return None

    def _looks_like_workflow(self, value):
        if not isinstance(value, dict) or not value:
            return False
        for node in value.values():
            if isinstance(node, dict) and str(node.get("class_type", "")).strip():
                return True
        return False

    def templates_for_context(self, source_file=None):
        templates = self.discover()
        media_type = ""
        if source_file:
            media_type = str(source_file.data.get("media_type", "")).strip().lower()

        filtered = []
        for template in templates:
            category = str(template.get("category", "unknown"))
            input_types = set(template.get("input_types", []))
            if source_file:
                if category not in ("enhance", "unknown"):
                    continue
                if input_types and media_type not in input_types:
                    continue
            else:
                if category not in ("create", "unknown"):
                    continue
                if category == "unknown" and input_types:
                    continue
            filtered.append(template)
        return filtered

    def get_template(self, template_id):
        template_id = str(template_id or "").strip()
        if not template_id:
            return None
        for template in self.discover():
            if str(template.get("id")) == template_id:
                return template
        return None

    def get_workflow_copy(self, template_id):
        template = self.get_template(template_id)
        if not template:
            return None
        return copy.deepcopy(template.get("workflow") or {})

    def output_icon_name(self, template):
        explicit_icon = str((template or {}).get("action_icon") or "").strip()
        if explicit_icon:
            return explicit_icon
        kind = str((template or {}).get("primary_output") or "unknown")
        if kind == "video":
            return "ai-action-create-video.svg"
        if kind == "audio":
            return "ai-action-create-audio.svg"
        if kind == "image":
            return "ai-action-create-image.svg"
        return "tool-generate-sparkle.svg"



"""
 @file
 @brief This file deals with unhandled exceptions
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import platform

from classes import info
from classes import sentry


def tail_file(f, n, offset=None):
    """Read the end of a file (n number of lines)"""
    avg_line_length = 90
    to_read = n + (offset or 0)

    while True:
        try:
            # Seek to byte position
            f.seek(-(avg_line_length * to_read), 2)
        except IOError:
            # Byte position not found
            f.seek(0)
        pos = f.tell()
        lines = f.read().splitlines()
        if len(lines) >= to_read or pos == 0:
            # Return the lines
            return lines[-to_read:offset and -offset or None]
        avg_line_length *= 2


def libopenshot_crash_recovery():
    """Walk libopenshot.log for the last line before this launch"""
    from classes.metrics import track_metric_error

    log_path = os.path.join(info.USER_PATH, "libopenshot.log")
    last_log_line = ""
    last_stack_trace = ""
    found_stack = False
    log_start_counter = 0
    if not os.path.exists(log_path):
        return
    with open(log_path, "rb") as f:
        # Read from bottom up
        for raw_line in reversed(tail_file(f, 500)):
            # Format and remove extra spaces from line
            line = " ".join(str(raw_line, 'utf-8').split()) + "\n"
            # Detect stack trace
            if "End of Stack Trace" in line:
                found_stack = True
                continue
            if "Unhandled Exception: Stack Trace" in line:
                found_stack = False
                continue
            if "libopenshot logging:" in line:
                log_start_counter += 1
                if log_start_counter > 1:
                    # Found the previous log start, too old now
                    break

            if found_stack:
                # Append line to beginning of stacktrace
                last_stack_trace = line + last_stack_trace

            # Ignore certain useless lines
            line.strip()
            if all(["---" not in line,
                    "libopenshot logging:" not in line,
                    not last_log_line,
                    ]):
                last_log_line = line

    # Split last stack trace (if any)
    if last_stack_trace:
        # Get top line of stack trace (for metrics)
        exception_lines = last_stack_trace.split("\n")
        last_log_line = exception_lines[0].strip()

        # Format and add exception log to Sentry context
        # Split list of lines into smaller lists (so we don't exceed Sentry limits)
        sentry.set_context("libopenshot", {"stack-trace": exception_lines})
        sentry.set_tag("component", "libopenshot")

    # Clear / normalize log line (so we can roll them up in the analytics)
    if last_log_line:
        # Format last log line based on OS (since each OS can be formatted differently)
        if platform.system() == "Darwin":
            last_log_line = "mac-%s" % last_log_line[58:].strip()
        elif platform.system() == "Windows":
            last_log_line = "windows-%s" % last_log_line
        elif platform.system() == "Linux":
            last_log_line = "linux-%s" % last_log_line.replace("/usr/local/lib/", "")

        # Remove '()' from line, and split. Trying to grab the beginning of the log line.
        last_log_line = last_log_line.replace("()", "")
        log_parts = last_log_line.split("(")
        if len(log_parts) == 2:
            last_log_line = "-%s" % log_parts[0].replace(
                "logger_libopenshot:INFO ", "").strip()[:64]
        elif len(log_parts) >= 3:
            last_log_line = "-%s (%s" % (log_parts[0].replace(
                "logger_libopenshot:INFO ", "").strip()[:64], log_parts[1])
    else:
        last_log_line = ""

    # Report exception (with last libopenshot line... if found)
    track_metric_error("unhandled-crash%s" % last_log_line, True)

    return last_log_line



"""
 @file
 @brief This file contains a theme's colors and UI dimensions
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2024 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QTabWidget, QWidget

from classes.info import PATH
from ..base import BaseTheme


class CosmicTheme(BaseTheme):
    def __init__(self, app):
        super().__init__(app)

        from classes.app import get_app
        _ = get_app()._tr

        self.style_sheet = """
QMainWindow {
    background-color: #192332;
    color: #91C3FF;
}

QWidget#tutorial {
    background-color: #192332;
    border: 1.2px solid #0078FF;
    border-radius: 4px;
    padding: 20px;
}

QLabel#lblTutorialText {
    font-size: 14px;
}

QCheckBox#checkboxMetrics {
    font-size: 14px;
}

QWidget#tutorial QPushButton#NextTip {
    background-color: #283241;
    font-size: 12px;
}

QWidget#tutorial QPushButton#HideTutorial {
    font-size: 12px;
}


QDialog {
    background-color: #192332;
    color: #91C3FF;
}

QLabel#lblMissingFileHint,
QLabel#lblMissingFilePath {
    color: #9bb2cc;
}

QWidget#Simple, QWidget#Advanced, QWidget#PreferencePanel {
    background-color: #141923;
    border: none;
}

QScrollArea {
    border: none;
}

QTabWidget {
    border: none;
}

QMenuBar {
    background-color: #283241;
    color: #91C3FF;
    padding: 0px;
    border: none;
}

QMenuBar::item {
    padding: 6px 10px;
    background: transparent;
}

QMenuBar::item:selected {
    background-color: #323C50;
    color: #ffffff;
}

QMenu {
    background-color: #141923;
    color: #91C3FF;
    padding: 6px;
    padding-left: 8px;
    padding-right: 8px;
    min-width: 40px;
    border: 1.2px solid #0078FF;
    border-radius: 3px 3px;
}

QMenu::item {
    padding: 6px 14px 6px 10px;
}

QMenu::item::checked {
    padding: 6px 12px 6px 20px;
}

QMenu::item:selected {
    background-color: #192332;
    color: #ffffff;
}

QMenu::separator {
    height: 8px;
    background-color: #141923;
}

QToolBar#toolBar {
    background-color: #283241;
    spacing: 0px;
    padding: 0px;
    border: none;
}

QToolBar#toolBar QToolButton {
    background-color: #283241;
    color: #91C3FF;
    padding-top: 10px;
    padding-bottom: 10px;
    padding-left: 8px;
    padding-right: 8px;
    border: none;
}

QToolBar#toolBar QToolButton:hover {
    background-color: #323C50;
}

QToolBar QToolButton:hover {
    background-color: #192332;
}

QToolBar QToolButton:pressed {
    background-color: #192332;
}

QToolBar#filesToolbar,
QToolBar#transitionsToolbar,
QToolBar#effectsToolbar {
    background-color: transparent;
    spacing: 4px;
    padding: 2px;
    border: none;
}

QToolBar#filesToolbar QToolButton,
QToolBar#transitionsToolbar QToolButton,
QToolBar#effectsToolbar QToolButton {
    background-color: #141923;
    color: #91C3FF;
    border-radius: 3px;
    padding: 3px 6px;
}

QToolBar#filesToolbar QToolButton:hover,
QToolBar#transitionsToolbar QToolButton:hover,
QToolBar#effectsToolbar QToolButton:hover {
    background-color: #192332;
}

QToolBar#filesToolbar QToolButton:checked,
QToolBar#transitionsToolbar QToolButton:checked,
QToolBar#effectsToolbar QToolButton:checked {
    background-color: #202b3a;
}

QToolBar#filesToolbar QToolButton:focus,
QToolBar#transitionsToolbar QToolButton:focus,
QToolBar#effectsToolbar QToolButton:focus {
    background-color: #1d2737;
}


QToolBar#timelineToolbar {
    background-color: #192332;
    spacing: 0px;
    padding: 0px;
    border: none;
}

QToolBar#timelineToolbar QToolButton {
    color: #91C3FF;
    background-color: #141923;
    padding: 8px;
    margin-bottom: 4px;
    margin-right: 5px;
    margin-left: 5px;
    border-radius: 4px;
    border: none;
}

QToolBar#timelineToolbar QToolButton:hover {
    background-color: #283241;
}


QToolBar#timelineToolbar QToolButton:checked {
    background-color: #283241;
}

QToolBar#toolBar QToolButton:focus {
    background-color: #2f3848;
}

QToolBar#timelineToolbar QToolButton:focus:!checked {
    background-color: #2a3444;
}

QToolBar#timelineToolbar QToolButton:checked:focus {
    background-color: #314055;
}

QToolBar#timelineToolbar QToolButton:pressed {
    background-color: #3a4558;
}

QToolBar#toolBar QToolButton:pressed {
    background-color: #3a4456;
}

QToolBar#filesToolbar QToolButton:pressed,
QToolBar#transitionsToolbar QToolButton:pressed,
QToolBar#effectsToolbar QToolButton:pressed {
    background-color: #2a374a;
}

QToolBar#toolBar QToolButton:hover {
    background-color: #323C50;
}

QToolBar#videoToolbar QToolButton:focus {
    background-color: #1d2737;
}

QToolBar#videoToolbar QToolButton:pressed {
    background-color: #2a374a;
}

QPushButton#acceptButton {
    padding: 8px 16px 8px 12px;
    border-radius: 4px;
    background-color: #0078FF;
    color: #FFFFFF;
}

QPushButton#acceptButton:hover {
    background-color: #006EE6;
}

QPushButton#acceptButton:focus,
QPushButton#acceptButton:default {
    background-color: #1a86ff;
}

QPushButton {
    padding: 8px 16px 8px 12px;
    border-radius: 4px;
    background-color: #192332;
    color: #91C3FF;
}

QPushButton:hover {
    background-color: #283241;
}

QPushButton:focus {
    background-color: #283241;
}

QWidget#settingsContainer {
    background-color: #141923;
}

QWidget#scrollAreaWidgetContents {
    background-color: #141923;
}

QPushButton#dock-close-button {
    image: url({PATH}themes/cosmic/images/dock-close.svg);
    padding: 0px;
    padding-top: 2px;
    padding-bottom: 2px;
    margin: 0px;
    margin-right: 16px;
    width: 16px;
    height: 16px;
}
QPushButton#dock-float-button {
    image: url({PATH}themes/cosmic/images/dock-float.svg);
    padding: 0px;
    padding-top: 2px;
    padding-bottom: 2px;
    margin: 0px;
    width: 16px;
    height: 16px;
}

QLabel#dock-title-label {
    color: #91C3FF;
    font-weight: 500;
    padding: 16px;
}

QLabel#dock-title-handle {
    padding-left: 16px;
    qproperty-pixmap: url({PATH}themes/cosmic/images/dock-move.svg);
}

QDockWidget {
    background-color: #141923;
    titlebar-close-icon: url({PATH}themes/cosmic/images/dock-close.svg);
    titlebar-normal-icon: url({PATH}themes/cosmic/images/dock-float.svg);
    color: #91C3FF;
    font-weight: 500;
    padding: 16px;
}

QDockWidget QWidget {
    border: none;
}

QDockWidget QWidget#dockFilesContents, QWidget#dockTransitionsContents, QWidget#dockEmojisContents, QWidget#dockEffectsContents, QWidget#dockCaptionContents, QWidget#dockVideoContents, QWidget#dockPropertiesContents {
    background-color: #141923;
    border-radius: 4px;
    margin-left: 16px;
    margin-right: 16px;
}

QDockWidget QWidget#dockTimelineContents {
    border-radius: 0px;
    margin-left: 0px;
    margin-right: 0px;
    padding: 0px;
}

QTabBar {
    border: none;
    qproperty-drawBase: 0;
    margin: 0px;
    padding: 0px;
}

QTabBar::tab {
    height: 16px;
    border: none;
    margin-left: 16px;
    margin-top: 16px;
    margin-bottom: 0px;
    padding-bottom: 0px;
    color: rgba(145, 195, 255, 0.4);
}

QTabWidget#exportTabs QTabBar::tab,
QTabWidget#tabCategories QTabBar::tab,
QTabWidget#tabCredits QTabBar::tab,
QTabWidget#generateTabs QTabBar::tab {
    margin-bottom: 10px;
}

QTabWidget#generateTabs QTabBar::tab:selected {
    border-bottom: 1.2px solid #53a0ed;
}

QTabBar::tab:selected {
    color: rgba(145, 195, 255, 1.0);
}

QTabBar:focus {
    outline: none;
}

QTabBar::tab:focus {
    border-bottom: 1.2px solid #53a0ed;
}

QToolBox::tab:focus {
    border-left: 1.2px solid #53a0ed;
}

QCheckBox:focus {
    background-color: #283241;
}

QLineEdit#filesFilter, QLineEdit#effectsFilter, QLineEdit#transitionsFilter, QLineEdit#emojisFilter, QLineEdit#txtPropertyFilter {
    background-color: #192332;
    border-radius: 4px;
    padding: 6px;
    padding-left: 8px;
    padding-right: 8px;
}

QLineEdit,
QSpinBox,
QDoubleSpinBox {
    padding-left: 6px;
}

QLineEdit#filesFilter:focus, QLineEdit#effectsFilter:focus, QLineEdit#transitionsFilter:focus, QLineEdit#emojisFilter:focus, QLineEdit#txtPropertyFilter:focus {
    border-width: 1.2px;
    border-style: solid;
    border-color: #53a0ed;
}

QScrollBar:vertical {
    border: none;
    background: transparent;
    width: 8px;
}

QScrollBar::handle:vertical {
    background: rgba(145, 195, 255, 51);
    border-radius: 4px;
    min-height: 20px;
}

QScrollBar::add-line:vertical, QScrollBar::sub-line:vertical {
    height: 0px;
    border: none;
    background: transparent;
}

QScrollBar::add-page:vertical, QScrollBar::sub-page:vertical {
    background: transparent;
}

QScrollBar:horizontal {
    border: none;
    background: transparent;
    height: 8px;
}

QScrollBar::handle:horizontal {
    background: rgba(145, 195, 255, 51);
    border-radius: 4px;
    min-width: 20px;
}

QScrollBar::add-line:horizontal, QScrollBar::sub-line:horizontal {
    width: 0px;
    border: none;
    background: transparent;
}

QScrollBar::add-page:horizontal, QScrollBar::sub-page:horizontal {
    background: transparent;
}

QComboBox {
    background-color: #121212;
    color: #FFFFFF;
    border: 1.2px solid transparent;
    border-radius: 4px;
    padding: 6px;
    padding-left: 8px;
    padding-right: 8px;
    combobox-popup: 0;
}

QComboBox:focus {
    border-color: #53a0ed;
}

QComboBox::drop-down {
    subcontrol-origin: padding;
    subcontrol-position: top right;
    width: 40px;
    border: none;
}

QComboBox::down-arrow {
    image: url({PATH}themes/cosmic/images/dropdown-arrow.svg);
}

QComboBox QAbstractItemView {
    color: #FFFFFF;
    border: 1.2px solid #0078FF;
    border-radius: 3px 3px 0px 0px;
    padding: 6px;
    padding-left: 8px;
    padding-right: 8px;
    background-color: #141923;
    text-align: left;
}

QComboBox::item {
    height: 24px;
}

QComboBox::item:selected {
    border: none;
    text-align: left;
    background-color: #192332;
}

QComboBox::item:checked {
    font-weight: bold;
    background-color: #192332;
}

QComboBox::indicator::checked {
    image: url({PATH}themes/cosmic/images/dropdown-tick.svg);
}

QHeaderView::section {
    background-color: #141923;
    color: #91C3FF;
    padding: 4px;
    border: none;
}

QTableView {
    background-color: #141923;
    gridline-color: #141923;
}

QTableView#propertyTableView::item:selected {
    background-color: #192332;
    border: 1.2px solid #0078FF;
}

QTreeView {
    background-color: #141923;
}

QListView {
    background-color: #141923;
}

QWidget#Details, QWidget#Output {
    background-color: #141923;
}

QWidget#Output QTextEdit {
    color: #91C3FF;
}

QToolBox::tab {
    color: #91C3FF;
    border-top: 1px solid rgba(145, 195, 255, .2);
}

QTabWidget QWidget#pageAdvancedOptions, QWidget#pageProfile, QWidget#pageImageSequenceSettings, QWidget#pageVideoSettings, QWidget#pageAudioSettings {
    background-color: #141923;
}

QTabWidget QWidget#pageVideoDetails, QWidget#pageVideoFormat, QWidget#pageAudioFormat, QWidget#pageFrameSettings {
    background-color: #141923;
}

QDialog#generateDialog QTabWidget#generateTabs::pane {
    border: none;
    background-color: #141923;
}

QDialog#generateDialog QTabWidget#generateTabs QWidget#pagePrompt,
QDialog#generateDialog QTabWidget#generateTabs QWidget#pagePoints,
QDialog#generateDialog QTabWidget#generateTabs QWidget#pageHighlight {
    background-color: #141923;
    border: none;
}

QDialog#generateDialog QLineEdit,
QDialog#generateDialog QTextEdit,
QDialog#generateDialog QComboBox {
    background-color: #141923;
    color: #91C3FF;
    border: 1.2px solid transparent;
    border-radius: 4px;
    padding: 6px 8px;
}

QDialog#generateDialog QLineEdit:focus,
QDialog#generateDialog QTextEdit:focus,
QDialog#generateDialog QComboBox:focus {
    border: 1.2px solid #53a0ed;
}

QLineEdit:disabled,
QTextEdit:disabled,
QComboBox:disabled,
QSpinBox:disabled,
QDoubleSpinBox:disabled {
    color: #808080;
}

QComboBox:disabled::drop-down {
    opacity: 0.75;
}

QWidget#cutting QPushButton#btnStart,QPushButton#btnEnd  {
    border: 1px solid #006EE6;
}

QWidget#cutting QPushButton#btnStart:disabled,QPushButton#btnEnd:disabled {
    color: #666666;
    border: 1px solid #666666;
}

QWidget#cutting QPushButton#btnAddClip {
    background-color: #006EE6;
    color: #FFFFFF;
}

QWidget#cutting QPushButton#btnAddClip:disabled {
    background-color: #283241;
    color: #666666;
}

.property_value {
    foreground-color: #0078FF;
    background-color: #283241;
}

.zoom_slider_playhead {
    background-color: #FABE0A;
}

QWidget#videoPreview {
    background-color: #141923;
}
        """
        path_unix_slashes = PATH.replace("\\", "/")
        self.style_sheet = f"""
QMessageBox QPushButton[text="&{_('Yes')}"] {{
    padding: 8px 16px 8px 12px;
    border-radius: 4px;
    background-color: #0078FF;
    color: #FFFFFF;
}}

QMessageBox QPushButton[text="&{_('Cancel')}"] {{
    qproperty-icon: none;
}}
        """ + self.style_sheet.replace("{PATH}", f"{path_unix_slashes}/")

    def apply_theme(self):
        super().apply_theme()

        from classes.app import get_app
        from classes import ui_util
        from classes.logger import log
        from PyQt5.QtWidgets import QStyleFactory
        from PyQt5.QtGui import QFont

        _ = get_app()._tr

        log.info("Setting Fusion dark palette")
        self.app.setStyle(QStyleFactory.create("Fusion"))
        dark_palette = ui_util.make_dark_palette(self.app.palette())
        self.app.setPalette(dark_palette)

        # Set font for all widgets
        font = QFont("Ubuntu")
        font.setPointSizeF(8)
        self.app.setFont(font)

        # Move tabs to top
        self.app.window.setTabPosition(Qt.TopDockWidgetArea, QTabWidget.North)

        # Set dock widget content margins to 0
        self.set_dock_margins([16, 0, 16, 0])
        self.set_dock_margins([0, 0, 0, 0], [0, 10, 0, 0], "dockTimelineContents")

        # Apply new stylesheet
        self.app.setStyleSheet(self.compose_stylesheet())

        # Create a transparent spacer widget
        spacer = QWidget(self.app.window)
        spacer.setFixedSize(15, 1)
        spacer.setStyleSheet("background: transparent;")

        # Main toolbar buttons
        toolbar_buttons = [
            {"action": self.app.window.actionNew, "icon": "themes/cosmic/images/tool-new-project.svg", "style": Qt.ToolButtonTextBesideIcon},
            {"action": self.app.window.actionOpen, "icon": "themes/cosmic/images/tool-open-project.svg", "style": Qt.ToolButtonTextBesideIcon},
            {"action": self.app.window.actionImportFiles, "icon": "themes/cosmic/images/tool-import-files.svg", "style": Qt.ToolButtonTextBesideIcon},
            {"action": self.app.window.actionProfile, "icon": "themes/cosmic/images/tool-profile.svg", "style": Qt.ToolButtonTextBesideIcon},
            {"expand": True},
            {"action": self.app.window.actionSave, "icon": "themes/cosmic/images/tool-save-project.svg", "style": Qt.ToolButtonTextBesideIcon},
            {"action": self.app.window.actionExportVideo, "icon": "themes/cosmic/images/tool-export.svg",
             "style": Qt.ToolButtonTextBesideIcon, "stylesheet": "QToolButton { background-color: #0078FF; color: #FFFFFF; border: none; } QToolButton:hover { background-color: #0a82ff; } QToolButton:focus { background-color: #0a82ff; } QToolButton:pressed { background-color: #3d9bff; }"},
            {"action": self.app.window.actionUpdate, "icon": "themes/cosmic/images/warning.svg", "visible": False, "style": Qt.ToolButtonTextBesideIcon, "stylesheet": "QToolButton {  background-color: #141923; color: #FABE0A; }"}
        ]
        self.set_toolbar_buttons(self.app.window.toolBar, icon_size=20, settings=toolbar_buttons)

        # Timeline toolbar buttons
        timeline_buttons = [
            {"action": self.app.window.actionAddTrack, "icon": "themes/cosmic/images/tool-add-track.svg", "style": Qt.ToolButtonTextBesideIcon, "stylesheet": "QToolButton { margin-left: 15px; }"},
            {"action": self.app.window.actionUndo, "icon": "themes/cosmic/images/tool-undo.svg", "style": Qt.ToolButtonIconOnly, "stylesheet": "QToolButton { margin-right: 0px; border-bottom-right-radius: 0px; border-top-right-radius: 0px; }"},
            {"action": self.app.window.actionRedo, "icon": "themes/cosmic/images/tool-redo.svg", "style": Qt.ToolButtonIconOnly, "stylesheet": "QToolButton { margin-left: 0px; border-bottom-left-radius: 0px; border-top-left-radius: 0px; }"},
            {"action": self.app.window.actionSnappingTool, "icon": "themes/cosmic/images/tool-snapping.svg", "style": Qt.ToolButtonIconOnly, "stylesheet": "QToolButton { margin-right: 0px; border-bottom-right-radius: 0px; border-top-right-radius: 0px; } QToolButton:focus { border: 1px solid #5aa2e6; background-color: #141923; } QToolButton:checked:focus { border: 1px solid #5aa2e6; background-color: #283241; }"},
            {"action": self.app.window.actionTimingTool, "icon": "themes/cosmic/images/tool-timing.svg", "style": Qt.ToolButtonIconOnly, "stylesheet": "QToolButton { margin-left: 0px; margin-right: 0px; border-bottom-left-radius: 0px; border-top-left-radius: 0px; border-bottom-right-radius: 0px; border-top-right-radius: 0px; } QToolButton:focus { border: 1px solid #5aa2e6; background-color: #141923; } QToolButton:checked:focus { border: 1px solid #5aa2e6; background-color: #283241; }"},
            {"action": self.app.window.actionRazorTool, "icon": "themes/cosmic/images/tool-razor.svg", "style": Qt.ToolButtonIconOnly, "stylesheet": "QToolButton { margin-left: 0px; border-bottom-left-radius: 0px; border-top-left-radius: 0px; } QToolButton:focus { border: 1px solid #5aa2e6; background-color: #141923; } QToolButton:checked:focus { border: 1px solid #5aa2e6; background-color: #283241; }"},
            {"action": self.app.window.actionAddMarker, "icon": "themes/cosmic/images/tool-add-marker.svg", "style": Qt.ToolButtonTextBesideIcon, "stylesheet": "QToolButton { margin-right: 0px; border-bottom-right-radius: 0px; border-top-right-radius: 0px; }"},
            {"action": self.app.window.actionPreviousMarker, "icon": "themes/cosmic/images/tool-prev-marker.svg", "style": Qt.ToolButtonIconOnly, "stylesheet": "QToolButton { margin-left: 0px; border-bottom-left-radius: 0px; border-top-left-radius: 0px; margin-right: 0px; border-bottom-right-radius: 0px; border-top-right-radius: 0px; }"},
            {"action": self.app.window.actionNextMarker, "icon": "themes/cosmic/images/tool-next-marker.svg", "style": Qt.ToolButtonIconOnly, "stylesheet": "QToolButton { margin-left: 0px; border-bottom-left-radius: 0px; border-top-left-radius: 0px; }"},
            {"action": self.app.window.actionCenterOnPlayhead, "icon": "themes/cosmic/images/tool-center-playhead.svg", "style": Qt.ToolButtonIconOnly, "stylesheet": "QWidget { margin-right: 10px; }"},
            {"widget": self.app.window.sliderZoomWidget},
            {"widget": spacer}
        ]
        self.set_toolbar_buttons(self.app.window.timelineToolbar, icon_size=12, settings=timeline_buttons)

        # Video toolbar
        toolbar_buttons = [
            {"expand": True},
            {"action": self.app.window.actionJumpStart, "icon": "themes/cosmic/images/tool-media-skip-back.svg", "style": Qt.ToolButtonIconOnly},
            {"action": self.app.window.actionRewind, "icon": "themes/cosmic/images/tool-media-rewind.svg", "style": Qt.ToolButtonIconOnly},
            {"action": self.app.window.actionPlay, "icon": "themes/cosmic/images/tool-media-play.svg", "style": Qt.ToolButtonIconOnly},
            {"action": self.app.window.actionFastForward, "icon": "themes/cosmic/images/tool-media-forward.svg", "style": Qt.ToolButtonIconOnly},
            {"action": self.app.window.actionJumpEnd, "icon": "themes/cosmic/images/tool-media-skip-forward.svg", "style": Qt.ToolButtonIconOnly},
            {"expand": True}
        ]
        self.set_toolbar_buttons(self.app.window.videoToolbar, icon_size=32, settings=toolbar_buttons)

        # Apply timeline theme
        self.app.window.timeline.apply_theme("""
            body {
              background: #141923;
            }
            ::-webkit-scrollbar {
              width: 8px;
              height: 8px;
            }
            ::-webkit-scrollbar-thumb {
                background: rgba(145, 195, 255, .2);
                border-radius: 4px;
            }
            ::-webkit-scrollbar-track {
              background: #141923;
              box-shadow: none;
            }
            .track_name {
              background: #192332;
              border-top-left-radius: 0px;
              border-bottom-left-radius: 0px;
              border: 1px solid #192332;
              border-left: 4px solid #0078FF;
              box-shadow: none;
              margin-left: 0px;
              height: 48px;
            }
            .track_top {
              padding-top: 4px;
              padding-left: 10px;
              background: none;
            }
            .track_label {
              text-shadow: none;
            }
            .track {
              background: #283241;
              border: 1px solid #283241;
              border-radius: 0px;
              height: 48px;
            }
            .track-resize-handle {
              background-color: #1B222CFF;
              border-top: 1px solid #1B222CFF;
              border-bottom: 1px solid #1B222CFF;
              border-right: 1px solid #1B222CFF;
              border-top-right-radius: 0px;
              border-bottom-right-radius: 0px;
            }
            .track-resize-handle:hover {
              background-color: #333F51FF;
            }
            .transition {
              height: 48px;
              min-height: 48px;
            }
            .transition_top {
              background: none;
            }
            .clip {
              background: #192332;
              border: 1px solid #0078FF;
              box-sizing: border-box;
              height: 48px;
              min-height: 48px;
              display: flex;
            }
            .clip_label {
              text-shadow: 1px 1px 1px black;
            }
            .clip_top {
              background: none;
              position: absolute;
              top: 0;
              width: 100%;
              display: flex;
              justify-content: flex-start;
              align-items: center;
              z-index: 2;
            }
            .thumb-container {
              margin-top: 0px;
            }
            .audio-container {
              margin-top: 0px;
              object-fit: cover;
              align-self: flex-start;
              z-index: 0;
            }
            .thumb-container, .audio-container {
              position: absolute;
              top: 0;
              left: 0;
              right: 0;
              bottom: 0;
              overflow: hidden;
            }
            .audio {
              border-radius: 8px;
            }
            .thumb {
              border-radius: 8px;
              margin: 0px;
              height: 100%;
              object-fit: cover;
              align-self: flex-start;
              z-index: 1;
            }
            .ui-selecting {
              border: 1px solid yellow !important;
            }
            .ui-selected {
              border: 1px solid red !important;
            }
            .playhead-top {
              margin-left: -6px;
              margin-top: 20px;
              width: 12px;
              height: 188px;
              background-image: url(../themes/cosmic/images/playhead.svg);
            }
            .playhead-line {
              z-index: 9999;
              position: absolute;
              top: 0;
              width: 2px;
              background-color: #FABE0A;
              margin: -1px;
              pointer-events: none;
            }
            .point_bezier {
              background-image: url(../themes/cosmic/images/keyframe-bezier.svg);
            }
            .point_linear {
              background-image: url(../themes/cosmic/images/keyframe-linear.svg);
            }
            .point_constant {
              background-image: url(../themes/cosmic/images/keyframe-constant.svg);
            }
            .track-keyframe-panel-disabled {
              background-image: url(../themes/cosmic/images/track-keyframe-panel-show-disabled.svg);
            }
            .track-keyframe-panel-enabled {
              background-image: url(../themes/cosmic/images/track-keyframe-panel-show-enabled.svg);
            }
            .track-add-above-disabled {
              background-image: url(../themes/cosmic/images/track-add-above-disabled.svg);
            }
            .track-add-above-enabled {
              background-image: url(../themes/cosmic/images/track-add-above-enabled.svg);
            }
            .track-add-below-disabled {
              background-image: url(../themes/cosmic/images/track-add-below-disabled.svg);
            }
            .track-add-below-enabled {
              background-image: url(../themes/cosmic/images/track-add-below-enabled.svg);
            }
            .track-delete-disabled {
              background-image: url(../themes/cosmic/images/track-delete-disabled.svg);
            }
            .track-delete-enabled {
              background-image: url(../themes/cosmic/images/track-delete-enabled.svg);
            }
            .track-locked-disabled {
              background-image: url(../themes/cosmic/images/track-locked-disabled.svg);
            }
            .track-locked-enabled {
              background-image: url(../themes/cosmic/images/track-locked-enabled.svg);
            }
            .track-unlocked-disabled {
              background-image: url(../themes/cosmic/images/track-unlocked-disabled.svg);
            }
            .track-unlocked-enabled {
              background-image: url(../themes/cosmic/images/track-unlocked-enabled.svg);
            }
            .keyframe-panel-add {
              background-image: url(../themes/cosmic/images/keyframe-panel-add.svg);
            }
            .marker_icon {
              background-image: url(../themes/cosmic/images/marker.svg);
            }
            #ruler_label {
              background: #141923;
            }
            #scrolling_ruler {
              background: #141923;
            }
        """)

        # Emit signal
        self.app.window.ThemeChangedSignal.emit(self)

    def togglePlayIcon(self, isPlay):
        """ Toggle the play icon from play to pause and back """
        button = self.app.window.videoToolbar.widgetForAction(self.app.window.actionPlay)
        if button:
            if not isPlay:
                play_icon_path = os.path.join(PATH, "themes/cosmic/images/tool-media-play.svg")
                button.setIcon(QIcon(play_icon_path))
            else:
                pause_icon_path = os.path.join(PATH, "themes/cosmic/images/tool-media-pause.svg")
                button.setIcon(QIcon(pause_icon_path))



"""
 @file
 @brief Unit tests for project asset path resolution
"""

import os
import sys
import unittest
from unittest.mock import patch


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from classes.assets import get_assets_path


class AssetsPathTests(unittest.TestCase):
    def test_get_assets_path_uses_user_path_for_backup_project(self):
        with patch("classes.assets.info.USER_PATH", "/home/test/.openshot_qt"), \
             patch("classes.assets.info.BACKUP_FILE", "/home/test/.openshot_qt/backup.osp"), \
             patch("classes.assets.info.RECOVERY_PATH", "/home/test/.openshot_qt/recovery"):
            asset_path = get_assets_path("/home/test/.openshot_qt/backup.osp", create_paths=False)

        self.assertEqual(asset_path, "/home/test/.openshot_qt")

    def test_get_assets_path_uses_user_path_for_recovery_project(self):
        with patch("classes.assets.info.USER_PATH", "/home/test/.openshot_qt"), \
             patch("classes.assets.info.BACKUP_FILE", "/home/test/.openshot_qt/backup.osp"), \
             patch("classes.assets.info.RECOVERY_PATH", "/home/test/.openshot_qt/recovery"):
            asset_path = get_assets_path("/home/test/.openshot_qt/recovery/123/project.osp", create_paths=False)

        self.assertEqual(asset_path, "/home/test/.openshot_qt")



"""
 @file
 @brief This file deals with value conversions
 @author Jonathan Thomas <jonathan@openshot.org>
 @author Frank Dana <ferdnyc AT gmail com>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

zoomSeconds = [1, 3, 5, 10, 15, 20, 30, 45, 60, 75,
               90, 120, 150, 180, 240, 300, 360, 480, 600, 720,
               900, 1200, 1500, 1800, 2400, 2700, 3600, 4800, 6000, 7200]

def zoomToSeconds(zoomValue):
    """ Convert zoom factor (slider position) into scale-seconds """
    if zoomValue < len(zoomSeconds):
        return zoomSeconds[zoomValue]
    else:
        return zoomSeconds[-1]

def secondsToZoom(scaleValue):
    """ Convert a number of seconds to a timeline zoom factor """
    if scaleValue in zoomSeconds:
        return zoomSeconds.index(scaleValue)
    else:
        # Find closest zoom
        closestValue = zoomSeconds[0]
        for zoomValue in zoomSeconds:
            if zoomValue < scaleValue:
                closestValue = zoomValue
        return zoomSeconds.index(closestValue)



"""
 @file
 @brief This file is for legacy support of OpenShot 1.x project files
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import uuid
from classes.legacy.openshot.classes import keyframe


class clip:
    """This class represents a media clip on the timeline."""

    # ----------------------------------------------------------------------
    def __init__(self, clip_name, color, position_on_track, start_time, end_time, parent_track, file_object):
        """Constructor"""

        # init variables for clip object
        self.name = clip_name  # the name of the clip
        self.color = color  # the color of the clip, used to organize clips
        self.start_time = start_time  # the time in seconds where we start playing a clip
        self.end_time = end_time  # the time in seconds where we stop playing a clip
        self.speed = 1.0  # the rate of playback (this will change if you want to slow down or speed up the clip)
        self.max_length = 0.0  # this is the max length of the clip in seconds
        self.position_on_track = float(
            position_on_track)  # the time in seconds to start playing the clip relative to the track
        self.play_video = True
        self.play_audio = True
        self.fill = True
        self.distort = False
        self.composite = True
        self.halign = "centre"
        self.valign = "centre"
        self.reversed = False
        self.volume = 100.0
        self.audio_fade_in = False
        self.audio_fade_out = False
        self.audio_fade_in_amount = 2.0
        self.audio_fade_out_amount = 2.0
        self.video_fade_in = False
        self.video_fade_out = False
        self.video_fade_in_amount = 2.0
        self.video_fade_out_amount = 2.0
        self.parent = parent_track  # the parent track this clip lives on
        self.file_object = file_object  # the file object that this clip is linked to
        self.unique_id = str(uuid.uuid1())
        self.rotation = 0.0
        self.thumb_location = ""

        # init key-frame dictionary
        self.keyframes = {"start": keyframe(0, 100.0, 100.0, 0.0, 0.0, 1.0),
                          "end": keyframe(-1, 100.0, 100.0, 0.0, 0.0, 1.0)}

        # init effects dictionary
        self.effects = []

        # init vars for drag n drop
        self.drag_x = 0.0
        self.drag_y = 0.0
        self.moved = False
        self.is_timeline_scrolling = False



"""
 @file
 @brief Optimize Preview submenu helpers for project file context menus.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os

from PyQt5.QtGui import QIcon

from classes import info
from classes.app import get_app
from .menu import StyledContextMenu


_OPTIMIZE_PREVIEW_READY_ICON = "tool-optimize-preview.svg"
_OPTIMIZE_PREVIEW_MISSING_ICON = "tool-optimize-preview-missing.svg"


def optimized_preview_icon(state="ready"):
     icon_name = _OPTIMIZE_PREVIEW_MISSING_ICON if state == "missing" else _OPTIMIZE_PREVIEW_READY_ICON
     icon_path = os.path.join(info.PATH, "themes", "cosmic", "images", icon_name)
     if os.path.exists(icon_path):
         return QIcon(icon_path)
     return QIcon()


def _optimizable_files(win):
     files = [f for f in (win.selected_files() or []) if f]
     eligible = []
     for file_obj in files:
         data = getattr(file_obj, "data", {}) or {}
         media_type = str(data.get("media_type", "") or "").strip().lower()
         if media_type == "video":
             eligible.append(file_obj)
     return eligible


def populate_optimized_preview_menu(win, optimized_menu):
     """Populate an Optimize Preview submenu for the current file selection."""
     optimized_menu.clear()
     selected_files = _optimizable_files(win)
     win._optimized_preview_target_file_ids = [str(getattr(f, "id", "") or "") for f in selected_files if getattr(f, "id", None)]
     if not selected_files:
         return None

     _ = get_app()._tr
     service = getattr(win, "proxy_service", None)
     states = [service.get_proxy_state(file_obj) for file_obj in selected_files] if service else []
     has_active = any(state in ("queued", "running", "canceling") for state in states)
     has_proxy = bool(service and any(service.has_proxy_reader(file_obj) for file_obj in selected_files))
     has_missing = "missing" in states
     use_existing_enabled = bool(selected_files) and not has_active
     win.actionOptimizedPreviewCreate.setText(_("Optimize Video"))
     win.actionOptimizedPreviewCreate.setEnabled(not has_active)
     win.actionOptimizedPreviewUseExisting.setEnabled(use_existing_enabled)
     win.actionOptimizedPreviewRemove.setEnabled(has_proxy)
     win.actionOptimizedPreviewCancel.setEnabled(has_active)
     win.actionOptimizedPreviewDeleteAndUnlink.setEnabled(has_proxy and not has_active)

     optimized_menu.setIcon(optimized_preview_icon("missing" if has_missing and not has_active else "ready"))
     if has_active:
         optimized_menu.addAction(win.actionOptimizedPreviewCancel)
         return optimized_menu
     if not has_active:
         optimized_menu.addAction(win.actionOptimizedPreviewCreate)
     optimized_menu.addAction(win.actionOptimizedPreviewUseExisting)
     if has_proxy:
         optimized_menu.addAction(win.actionOptimizedPreviewRemove)
         optimized_menu.addAction(win.actionOptimizedPreviewDeleteAndUnlink)
     return optimized_menu


def add_optimized_preview_menu(win, menu):
     """Add the Optimize Preview submenu for the current file selection."""
     selected_files = _optimizable_files(win)
     if not selected_files:
         return None

     optimized_menu = StyledContextMenu(title=get_app()._tr("Optimize"), parent=menu)
     populate_optimized_preview_menu(win, optimized_menu)
     menu.addMenu(optimized_menu)
     return optimized_menu



"""
 @file
 @brief This file manages the optional Sentry SDK
 @author Jonathan Thomas <jonathan@openshot.org>
 @author FeRD (Frank Dana) <ferdnyc@gmail.com>

 @section LICENSE

 Copyright (c) 2008-2021 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import platform
import datetime
from classes.logger import log

from classes import info
from classes.logger import log

try:
    import distro
except ModuleNotFoundError:
    distro = None

try:
    import sentry_sdk as sdk
except ModuleNotFoundError:
    sdk = None

# seconds required between errors
min_error_freq = 1
last_send_time = None
last_event_message = None

def init_tracing():
    """Init all Sentry tracing"""
    if not sdk:
        log.info('No sentry_sdk module detected (error reporting is disabled)')
        return

    # Determine sample rate for errors & transactions
    sample_rate = 0.0
    traces_sample_rate = 0.0
    if info.VERSION == info.ERROR_REPORT_STABLE_VERSION:
        sample_rate = info.ERROR_REPORT_RATE_STABLE
        traces_sample_rate = info.TRANS_REPORT_RATE_STABLE
        environment = "production"
    else:
        sample_rate = info.ERROR_REPORT_RATE_UNSTABLE
        traces_sample_rate = info.TRANS_REPORT_RATE_UNSTABLE
        environment = "unstable"

    if info.ERROR_REPORT_STABLE_VERSION:
        log.info("Sentry initialized for '%s': %s sample rate, %s transaction rate" % (environment,
                                                                                       sample_rate,
                                                                                       traces_sample_rate))

    def before_send(event,hint):
        """
        Function to filter out repetitive Sentry.io errors before sending them
        """
        global last_send_time
        global last_event_message

        # Prevent rapid errors
        current_time = datetime.datetime.now()
        if last_send_time:
            time_since_send = (current_time - last_send_time).total_seconds()
            if time_since_send < min_error_freq:
                log.debug("Report prevented: Recent error reported")
                return None

        # Prevent repeated errors
        event_message = event.\
            get("logentry", {"message": None}).\
            get("message", None)
        if last_event_message and last_event_message == event_message:
            log.debug("Report prevented: Same as last Error")
            return None

        # This error will send. Update the last time and last message
        log.debug("Sending Error")
        last_send_time = current_time
        last_event_message = event_message
        return event

    # Sentry is optional. If transport setup fails due to a broken host
    # environment (for example malformed proxy variables), continue startup
    # without telemetry instead of crashing the entire application.
    try:
        sdk.init(
            "https://21496af56ab24e94af8ff9771fbc1600@o772439.ingest.sentry.io/5795985",
            sample_rate=sample_rate,
            traces_sample_rate=traces_sample_rate,
            release=f"openshot@{info.VERSION}",
            environment=environment,
            debug=False,
            before_send=before_send
        )
    except Exception as ex:
        log.warning("Sentry initialization failed, disabling error reporting: %s", ex)
        return
    if _supports_tagging():
        configure_platform_tags(sdk)
    else:
        sdk.configure_scope(platform_scope)


def platform_scope(scope):
    configure_platform_tags(scope)


def configure_platform_tags(sdk_or_scope):
    sdk_or_scope.set_tag("system", platform.system())
    sdk_or_scope.set_tag("machine", platform.machine())
    sdk_or_scope.set_tag("processor", platform.processor())
    sdk_or_scope.set_tag("platform", platform.platform())
    if distro and platform.system() == "linux":
        sdk_or_scope.set_tag("distro", " ".join(distro.linux_distribution()))
    sdk_or_scope.set_tag("locale", info.CURRENT_LANGUAGE)


def disable_tracing():
    """Disable all Sentry tracing requests"""
    if sdk:
        sdk.init()


def set_tag(*args):
    if sdk and _supports_tagging():
        sdk.set_tag(*args)


def set_user(*args):
    if sdk and _supports_tagging():
        sdk.set_user(*args)


def set_context(*args):
    if sdk and _supports_tagging():
        sdk.set_context(*args)


def _supports_tagging():
    """Returns whether the imported sentry-sdk has tag-related
    methods such as set_tag, set_user, set_context.

    Those methods were introduce on 0.13.1 version. Checking this
    before calling those methods on the sentry-sdk avoids crashing
    Openshot in case an old sdk is installed.
    """
    return all([hasattr(sdk, method) for method in ["set_tag", "set_user", "set_context"]])
